#include "Cyclone.h"
#include "m68k.h"
#include <stdio.h>

int MC68000_ICount;
int pending_interrupts;

struct Cyclone cyclone;

static unsigned int MyCheckPc(unsigned int pc)
{
	pc -= cyclone.membase;
	if(pc<0x200000) {
		extern char *neogeo_prg_memory;
		cyclone.membase=(int)neogeo_prg_memory;
	} else {
		switch(pc>>16)
		{
			/* BIOS ROM */
			case    0xc0:
			case    0xc1:
			case    0xc2:
			case    0xc3:
			case    0xc4:
			case    0xc5:
			case    0xc6:
			case    0xc7:
			{
				extern char *neogeo_rom_memory;
				cyclone.membase=(int)neogeo_rom_memory - 0xc00000;
				break;
			}

			default:
				printf("error!! PC=%x\n", pc);
				break;
		}
	}
  	return cyclone.membase+pc; /* New program counter */
}

void m68k_pulse_reset(void)
{
	memset(&cyclone, 0,sizeof(cyclone));
	cyclone.checkpc=MyCheckPc;
	cyclone.read8  =m68k_read_memory_8;
	cyclone.read16 =m68k_read_memory_16;
	cyclone.read32 =m68k_read_memory_32;
	cyclone.write8 =m68k_write_memory_8;
	cyclone.write16=m68k_write_memory_16;
	cyclone.write32=m68k_write_memory_32;
	cyclone.fetch8  =m68k_read_memory_8;
	cyclone.fetch16 =m68k_read_memory_16;
	cyclone.fetch32 =m68k_read_memory_32;
  	cyclone.srh=0x27; /* Set supervisor mode */
	cyclone.irq=0x00;
  	cyclone.a[7]=cyclone.read32(0); /* Get Stack Pointer */
  	cyclone.membase=0;
  	cyclone.pc=cyclone.checkpc(cyclone.read32(4)); /* Get Program Counter */
   	pending_interrupts = 0;
}

unsigned int m68k_get_reg(void* context, m68k_register_t regnum)
{
	switch(regnum)
	{
		case M68K_REG_D0:	return cyclone.d[0];
		case M68K_REG_D1:	return cyclone.d[1];
		case M68K_REG_D2:	return cyclone.d[2];
		case M68K_REG_D3:	return cyclone.d[3];
		case M68K_REG_D4:	return cyclone.d[4];
		case M68K_REG_D5:	return cyclone.d[5];
		case M68K_REG_D6:	return cyclone.d[6];
		case M68K_REG_D7:	return cyclone.d[7];
		case M68K_REG_A0:	return cyclone.a[0];
		case M68K_REG_A1:	return cyclone.a[1];
		case M68K_REG_A2:	return cyclone.a[2];
		case M68K_REG_A3:	return cyclone.a[3];
		case M68K_REG_A4:	return cyclone.a[4];
		case M68K_REG_A5:	return cyclone.a[5];
		case M68K_REG_A6:	return cyclone.a[6];
		case M68K_REG_A7:	return cyclone.a[7];
		case M68K_REG_PC:	return cyclone.pc - cyclone.membase;
		default:			printf("m68k_get_reg Error!! %d.\n", regnum); return 0;
	}
	return 0;
}

void m68k_set_reg(m68k_register_t regnum, unsigned int value)
{
	switch(regnum)
	{
		case M68K_REG_A7:	cyclone.a[7] = value; return;
		case M68K_REG_PC:	cyclone.pc = cyclone.checkpc(value + cyclone.membase); return;
		case M68K_REG_SR:	cyclone.srh = value >> 8; return;
		case M68K_REG_USP:
		case M68K_REG_ISP:	cyclone.osp = value; return;
		default:			printf("m68k_set_reg Error!! %d, %x.\n", regnum, value); return;
	}
}


void MC68000_Clear_Pending_Interrupts(void)
{
	pending_interrupts &= ~0xff000000;
}


void m68k_set_irq(unsigned int level)
{
	if(level > cyclone.irq)
		cyclone.irq = level;
/*
	if (level >= 1 && level <= 7)
		pending_interrupts |= 1 << (24 + level);
*/
}

/* NeoCD declarations */
extern int img_display;

int m68k_execute(int cycles)
{
	//MC68000_ICount = cycles;
/*
	if (pending_interrupts & 0xff000000) {
		while ((MC68000_ICount > 0) && (pending_interrupts & 0xff000000))
		{
			int level, mask = 0x80000000;
			for (level = 7; level; level--, mask >>= 1)
				if (pending_interrupts & mask)
					break;
   			if(level>cyclone.irq)
   			{
   				pending_interrupts &= ~(1 << (level + 24));
   				cyclone.irq = level;
   			}
			cyclone.cycles=46;
			printf("Int Run PC=%x, IRQ=%d.\n", cyclone.pc - cyclone.membase, cyclone.irq);
			CycloneRun(&cyclone);
			MC68000_ICount-=(46-cyclone.cycles);
		}
	}

	if (MC68000_ICount > 0) {
		printf("Run PC=%x.\n", cyclone.pc - cyclone.membase);
		cyclone.cycles=MC68000_ICount;
		CycloneRun(&cyclone);
		MC68000_ICount = cyclone.cycles;
	}
*/

	cyclone.cycles=cycles;
	while(cyclone.cycles > 0) {
		CycloneRun(&cyclone);
		//printf("Last OpCode %4x, cycles=%d.\n", cyclone.lastest_opcode, cyclone.cycles);
		switch (cyclone.lastest_opcode) {
			case 0xfabe: neogeo_exit(); break;
			case 0xfabf: img_display=1; cdrom_load_files(); break;
			case 0xfac0: img_display=0; cdrom_load_files(); break;
			case 0xfac1: neogeo_upload(); break;
			case 0xfac2: neogeo_prio_switch(); break;
			case 0xfac3: neogeo_cdda_control(); break;
		}
	}

	//printf("PC=%x.(%d, %d)\n", cyclone.pc - cyclone.membase, cycles, cycles - cyclone.cycles);
   	return (cycles - cyclone.cycles);
}
