#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "MadWrapper.h"

extern volatile int cdda_autoloop;
extern int main_volume;

MAD_WRAPPER *mad = NULL;
SDL_RWops *mad_src = NULL;
int src_begin;
int MP3_playing = 0;
SDL_mutex *mad_mutex = NULL;
#define EnterSound() SDL_LockMutex(mad_mutex)
#define LeaveSound() SDL_UnlockMutex(mad_mutex)

int PlayMP3(char *name);
void StopMP3();

void MP3Update(Uint8 *stream, int len) {
	int mixlen, mixpos;

	if(!mad_mutex) return;
	EnterSound();
	if(!MP3_playing) {
		LeaveSound();
		return;
	}

	mixpos = 0;
	do {
		mixlen = MAD_WRAPPER_playAudio( mad, stream + mixpos, len - mixpos );
		mixpos += mixlen;
		if(mixlen == 0) {
			if(cdda_autoloop) {
				SDL_RWseek(mad_src, src_begin, SEEK_SET);
			} else {
				LeaveSound();
				StopMP3();
				EnterSound();
			}
		}
	} while (cdda_autoloop && (len < mixpos));
	LeaveSound();
}

void MP3Pause(int pause) {
	if(mad_mutex) {
		EnterSound();
		MP3_playing = !pause;
		LeaveSound();
	} else {
		MP3_playing = !pause;
	}
}

int PlayMP3(char *name) {
	if(!mad_mutex)
		mad_mutex = SDL_CreateMutex();

	StopMP3();
	printf("MP3 File: %s\n", name);

	mad_src = SDL_RWFromFile(name, "rb");
	if(mad_src) {
		src_begin = 0;
		mad = MAD_WRAPPER_new_rwops(mad_src, 0, 0);
		MAD_WRAPPER_setvolume(mad, main_volume);
		MAD_WRAPPER_play(mad);
		MP3_playing = 1;

		return 0;
	};

	return 1;
}

void StopMP3() {
	if(mad_mutex) {
		EnterSound();
		MP3_playing = 0;
		if(mad) {
			MAD_WRAPPER_delete(mad);
			mad_src = NULL;
			mad = NULL;
		}
		LeaveSound();
	} else {
		MP3_playing = 0;
		if(mad) {
			MAD_WRAPPER_delete(mad);
			mad_src = NULL;
			mad = NULL;
		}
	}
}

void DeinitMP3() {
	if(mad_mutex) {
		StopMP3();
		SDL_DestroyMutex(mad_mutex);
		mad_mutex = NULL;
	}
}

void SetVolumeMP3() {
	if(mad_mutex && mad) {
		EnterSound();
		if(main_volume < 0) main_volume = 0;
		if(main_volume > 100) main_volume = 100;
		MAD_WRAPPER_setvolume(mad, main_volume);
		LeaveSound();
	}
}
