#include  <stdio.h>
#include  <stdlib.h>
#include "neocd.h"
#include "sound/2610intf.h"
#include "z80intrf.h"

#ifndef INLINE
#define INLINE	static __inline
#endif

#define INT_IRQ 0x01
#define NMI_IRQ 0x02

/* Push PC -> SP-=2, PC->SP */
#define PUSH_PC() { drz80.Z80SP-=2; drz80.z80_write16(drz80.Z80PC - drz80.Z80PC_BASE,drz80.Z80SP); }

static struct DrZ80 drz80;

unsigned char subcpu_memspace[0x10000]__attribute__((aligned(32)));

unsigned int sound_code;
unsigned int result_code;
unsigned int pending_command;
int	Z80_VBL_CYCLES;
int	z80_cycles;
unsigned short z80_bank[4];

static unsigned int z80_rebasePC(unsigned short address)
{
	drz80.Z80PC_BASE = (unsigned int)subcpu_memspace;
	return drz80.Z80PC_BASE + address;
}

static unsigned int z80_rebaseSP(unsigned short address)
{
	drz80.Z80SP_BASE = (unsigned int)subcpu_memspace;
	return drz80.Z80SP_BASE + address;
}

static unsigned char z80_read8(unsigned short address) {
	return (subcpu_memspace[address&0xFFFF]);
}

static unsigned short z80_read16(unsigned short address) {
	return z80_read8(address) | (z80_read8(address + 1) << 8);
}

static void z80_write8(unsigned char data,unsigned short address) {
	subcpu_memspace[address&0xFFFF]=data;
}

static void z80_write16(unsigned short data,unsigned short address) {
	z80_write8(address, data & 0xFF);
	z80_write8(address + 1, data >> 8);
}

unsigned char z80_in(unsigned char p) {
	switch( p & 0xff) {
		case 0x0: pending_command = 0; return sound_code;
		case 0x4: return YM2610_status_port_0_A_r(0);
		case 0x5: return YM2610_read_port_0_r(0);
		case 0x6: return YM2610_status_port_0_B_r(0);
	};
	return 0;
}

static void z80_out(unsigned char p,unsigned char d) {
	switch( p & 0xff) {
	case 0x4: YM2610_control_port_0_A_w(0, d); break;
	case 0x5: YM2610_data_port_0_A_w(0, d); break;
	case 0x6: YM2610_control_port_0_B_w(0, d); break;
	case 0x7: YM2610_data_port_0_B_w(0, d); break;
	case 0xc: result_code = d; break;
/*
	case 0x8: // NMI enable / acknowledge? (the data written doesn't matter)
		// Metal Slug Passes this 35, then 0 in sequence. After a mission begins it passes it 1.
		break;
	case 0x18: // NMI disable? (the data written doesn't matter)
		break;
	case 0xc0: //cyberlip calls this for some reason, maybe other games?
		break;
	case 0xc1: //cyberlip calls this for some reason, maybe other games?
		break;
	case 0x00: //This is called by Aero Fighters 2 at startup as well as a few others.
		//They first disable the NMI, call this and pass 80, then enable the NMI again
		break;
	default:
		sprintf(z80_str,"Unimplemented Z80 Write Port: %x data: %x\n",PortNo&0xff,data);
		debug_log(z80_str);
		break;
*/
	}
}

static void z80_irq_callback(void)
{
	drz80.Z80_IRQ = 0x00;
}

/****************************************************************************
 * Reset registers to their initial values
 ****************************************************************************/
void z80_init(void)
{
	Z80_VBL_CYCLES = 4000000/FPS;
	z80_cycles=Z80_VBL_CYCLES;
	memset (&drz80, 0, sizeof(drz80));
  	drz80.z80_rebasePC=z80_rebasePC;
  	drz80.z80_rebaseSP=z80_rebaseSP;
  	drz80.z80_read8   =z80_read8;
  	drz80.z80_read16  =z80_read16;
  	drz80.z80_write8  =z80_write8;
  	drz80.z80_write16 =z80_write16;
  	drz80.z80_in      =z80_in;
  	drz80.z80_out     =z80_out;
  	drz80.z80_irq_callback=z80_irq_callback;
  	drz80.Z80A = 0x00 <<24;
  	drz80.Z80F = (1<<2); /* set ZFlag */
  	drz80.Z80BC = 0x0000	<<16;
  	drz80.Z80DE = 0x0000	<<16;
  	drz80.Z80HL = 0x0000	<<16;
  	drz80.Z80A2 = 0x00 <<24;
  	drz80.Z80F2 = 1<<2;  /* set ZFlag */
  	drz80.Z80BC2 = 0x0000 <<16;
  	drz80.Z80DE2 = 0x0000 <<16;
  	drz80.Z80HL2 = 0x0000 <<16;
  	drz80.Z80IX = 0xFFFF	<<16;
  	drz80.Z80IY = 0xFFFF	<<16;
	drz80.Z80I = 0x00;
  	drz80.Z80IM = 0x00;
  	drz80.Z80_IRQ = 0x00;
  	drz80.Z80IF = 0x00;
	//drz80.Z80PC_BASE = (unsigned int)subcpu_memspace;
  	drz80.Z80PC=drz80.z80_rebasePC(0);
	//drz80.Z80SP_BASE = (unsigned int)subcpu_memspace;
  	drz80.Z80SP=drz80.z80_rebaseSP(0xffff);
	Z80_Execute(100000);
}

void Z80_Interrupt(int type)
{
	if (type == Z80_NMI_INT) {
		drz80.Z80_IRQ = 0x00;
		drz80.Z80IF = (drz80.Z80IF&1)<<1;  /* Save interrupt flip-flop 1 to 2 and Clear interrupt flip-flop 1 */
		PUSH_PC();
		drz80.Z80PC=drz80.z80_rebasePC(0x0066); /* _PCD = 0x0066; */
	} else {
	    drz80.z80irqvector = 0xFF;  // default IRQ vector RST opcode
		drz80.Z80_IRQ = type;
	}
}

/****************************************************************************
 * Execute IPeriod T-states. Return number of T-states really executed
 ****************************************************************************/
int Z80_Execute(int cycles)
{
	DrZ80Run(&drz80, cycles);

	return cycles;
}
