/*===================================================================*/
/*                                                                   */
/*  InfoNES_System_SDL.cpp : SDL specific File                       */
/*                                                                   */
/*  2001/05/18  InfoNES Project ( Sound is based on DarcNES )        */
/*                                                                   */
/*===================================================================*/

/*-------------------------------------------------------------------*/
/*  Include files                                                    */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <SDL.h>

#include "../InfoNES.h"
#include "../InfoNES_System.h"
//#include "../InfoNES_pAPU.h"
#include "apu/nes_apu.h"
#include "filer.h"

#ifdef WIN32
#include <windows.h>
#else
#include <stdarg.h>
#include <unistd.h>
#endif

#define REFRESHTIME 1000/60

/*-------------------------------------------------------------------*/
/*  ROM image file information                                       */
/*-------------------------------------------------------------------*/

char szRomName[ _MAX_PATH ];
char szSaveName[ _MAX_PATH ];
int nSRAM_SaveFlag;

/*-------------------------------------------------------------------*/
/*  Constants ( Linux specific )                                     */
/*-------------------------------------------------------------------*/

#define VBOX_SIZE    7 
#define VERSION      "InfoNES v0.96J"

/*-------------------------------------------------------------------*/
/*  Global Variables ( Linux specific )                              */
/*-------------------------------------------------------------------*/

bool bRunning;
SDL_Surface *screen;
#ifdef GP2X
int joybtn[19];
bool joydown[19];
#endif

/* Pad state */
DWORD dwKeyPad1;
DWORD dwKeyPad2;
DWORD dwKeySystem;

/* For Sound Emulation */
BYTE final_wave[4096];
int waveptr;
int wavflag;
bool sound_on;
bool sound_open = false;
int main_volume;

static char work_path[_MAX_PATH];
static char save_path[_MAX_PATH];
static char roms_path[_MAX_PATH];
static char config_file[_MAX_PATH];

/*-------------------------------------------------------------------*/
/*  Function prototypes ( Linux specific )                           */
/*-------------------------------------------------------------------*/
void SDL_ProcessEvent( void );
void Finalize( void );
void Reset( void );

int LoadSRAM();
int SaveSRAM();

/* Palette data */
WORD NesPalette[ 64 ] =
{
	0x39ce, 0x1071, 0x0015, 0x2013, 0x440e, 0x5402, 0x5000, 0x3c20,
	0x20a0, 0x0100, 0x0140, 0x00e2, 0x0ceb, 0x0000, 0x0000, 0x0000,
	0x5ef7, 0x01dd, 0x10fd, 0x401e, 0x5c17, 0x700b, 0x6ca0, 0x6521,
	0x45c0, 0x0240, 0x02a0, 0x0247, 0x0211, 0x0000, 0x0000, 0x0000,
	0x7fff, 0x1eff, 0x2e5f, 0x223f, 0x79ff, 0x7dd6, 0x7dcc, 0x7e67,
	0x7ae7, 0x4342, 0x2769, 0x2ff3, 0x03bb, 0x0000, 0x0000, 0x0000,
	0x7fff, 0x579f, 0x635f, 0x6b3f, 0x7f1f, 0x7f1b, 0x7ef6, 0x7f75,
	0x7f94, 0x73f4, 0x57d7, 0x5bf9, 0x4ffe, 0x0000, 0x0000, 0x0000
};

/*===================================================================*/
/*                                                                   */
/*                main() : Application main                          */
/*                                                                   */
/*===================================================================*/

/* Application main */
int main(int argc, char **argv)
{                                  
	int skin;
	FILE *fp;

	/*-------------------------------------------------------------------*/
	/*  Initialize SDL                                                   */
	/*-------------------------------------------------------------------*/
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_AUDIO) != 0 ) 
	{
		printf("Could not initialize SDL(%s)\n", SDL_GetError());
		exit(0);
	}
	atexit(Finalize);

#ifdef GP2X
	SDL_ShowCursor(SDL_DISABLE);
	screen = SDL_SetVideoMode(320, 240, 16, SDL_FULLSCREEN);
#else
	screen = SDL_SetVideoMode(320, 240, 16, 0);
#endif
	if (screen == NULL)
	{
		printf("Couldn't set video mode: %s\n", SDL_GetError());
		exit(0);
	}

#ifdef GP2X
	// Check for joystick
	if(SDL_NumJoysticks()>0)
	{
		SDL_Joystick *joy = SDL_JoystickOpen(0);
		
		if(joy)
		{
			printf("Opened Joystick 0\n");
			printf("Name: %s\n", SDL_JoystickName(0));
			printf("Number of Axes: %d\n", SDL_JoystickNumAxes(joy));
			printf("Number of Buttons: %d\n", SDL_JoystickNumButtons(joy));
			printf("Number of Balls: %d\n", SDL_JoystickNumBalls(joy));

			memset(joybtn, 0, sizeof(joybtn));

			joybtn[0] = ( 1 << 4 );
			joybtn[4] = ( 1 << 5 );
			joybtn[2] = ( 1 << 6 );
			joybtn[6] = ( 1 << 7 );

			joybtn[1] = joybtn[0] | joybtn[2];
			joybtn[3] = joybtn[4] | joybtn[2];
			joybtn[5] = joybtn[4] | joybtn[6];
			joybtn[7] = joybtn[0] | joybtn[6];

			joybtn[0x08] = ( 1 << 3 ); // Start
			joybtn[0x09] = ( 1 << 2 ); // Select
#if 0
			joybtn[0x0C] = ( 1 << 1 ); // A
			joybtn[0x0E] = ( 1 << 1 ); // A
			joybtn[0x0D] = ( 1 << 0 ); // B
			joybtn[0x0F] = ( 1 << 0 ); // B
#elif 1 // original NES mapping
			joybtn[0x0D] = ( 1 << 1 ); // A
			joybtn[0x0F] = ( 1 << 1 ); // A
			joybtn[0x0C] = ( 1 << 0 ); // B
			joybtn[0x0E] = ( 1 << 0 ); // B
#endif

		}
		else
		{
			printf("Couldn't open Joystick 0\n");
		}
	}
#endif

	/*-------------------------------------------------------------------*/
	/*  Pad Control                                                      */
	/*-------------------------------------------------------------------*/

	/* Initialize a pad state */
	dwKeyPad1   = 0;
	dwKeyPad2   = 0;
	dwKeySystem = 0;

#ifdef __unix__
	getcwd(work_path, _MAX_PATH);
	strcat(work_path, "/");
#else
	int i;

	strcpy(work_path, argv[0]);
	for(i = strlen(work_path)-1; i>0 ; i--) {
		if((work_path[i-1] == '/') || (work_path[i-1] == '\\')) {
			work_path[i] = '\0';
			break;
		}
	}
#endif

	strcpy(save_path, work_path);
	strcat(save_path, "save/");

#ifdef WIN32
	CreateDirectory(save_path, NULL);
#else
	mkdir (save_path, 0777);
	chown (save_path, getuid (), getgid ());
#endif

	strcpy(roms_path, work_path);
	strcat(roms_path, "roms/");

	strcpy(config_file, save_path);
	strcat(config_file, "nesgp2x.cfg");

	main_volume = 80;

#ifdef __GP2X__
	system("/bin/mount -t vfat -o remount,sync,iocharset=utf8 /dev/mmcsd/disc0/part1 /mnt/sd");
#endif

	fp = fopen(config_file, "r");
	if(fp) {
		fscanf(fp, "%d", &skin);
		fclose(fp);
	} else
		skin = 0;
	if((skin < 0) || (skin > 9)) skin = 0;

	/*-------------------------------------------------------------------*/
	/*  Load Cassette & Create Thread                                    */
	/*-------------------------------------------------------------------*/

	/* If a rom name specified, start it */
	while ( 1 ) // argc == 2 )
	{
		/* Set a ROM image name */
		// Initialize End... (Filer Start)
		if(!filer(work_path, roms_path, szRomName, &skin)) {
			exit(0);
		} else {
			fp = fopen(config_file, "w");
			if(fp) {
				fprintf(fp, "%d\n", skin);
				fclose(fp);
			}
		}

		printf("Select Rom: %s\n", szRomName);

		/* Load cassette */
		if ( InfoNES_Load ( szRomName ) == 0 )
		{ 
			/* Load SRAM */
			LoadSRAM();
			bRunning = true;
			SDL_FillRect(screen, NULL, 0);
			InfoNES_Main();
			SaveSRAM();
		}
	}

	return(0);
}

/*===================================================================*/
/*                                                                   */
/*     SDL_ProcessEvent() : Connecting to the key_press_event event  */
/*                                                                   */
/*===================================================================*/
#ifdef GP2X
void Joystick( void )
{
	int allbtn = 0;
	int keys = 0;
	int button;
	char joyoff[19];

	memset(joyoff, 0, sizeof(joyoff));
	if(joydown[0] || joydown[2] || joydown[4] || joydown[6]) {
		joyoff[1] = joyoff[3] = joyoff[5] = joyoff[7] = 1;
	}

	for (button=0; button<19; button++) {
		if(!joyoff[button] && joydown[button])
			keys |= joybtn[button];
		allbtn |= joybtn[button];
	}
	dwKeyPad1 &= ~allbtn;
	dwKeyPad1 |= keys;
}
#endif

void SDL_ProcessEvent( void )
{
	SDL_Event event;
	while(SDL_PollEvent(&event)) {
		switch(event.type) {
		case SDL_KEYDOWN:
			switch ( event.key.keysym.sym )
			{
				case SDLK_RIGHT:
					dwKeyPad1 |= ( 1 << 7 );
					break;

				case SDLK_LEFT:
					dwKeyPad1 |= ( 1 << 6 );
					break;

				case SDLK_DOWN:
					dwKeyPad1 |= ( 1 << 5 );
					break;
					
				case SDLK_UP:
					dwKeyPad1 |= ( 1 << 4 );
					break;

				case 's':
				case 'S':
					/* Start */
					dwKeyPad1 |= ( 1 << 3 );
					break;

				case 'a':
				case 'A':
					/* Select */
					dwKeyPad1 |= ( 1 << 2 );
					break;

				case 'z':                     
				case 'Z':                     
					/* 'A' */
					dwKeyPad1 |= ( 1 << 1 );
					break;

				case 'x': 
				case 'X': 
					/* 'B' */
					dwKeyPad1 |= ( 1 << 0 );
					break;

				case 'c':
				case 'C':   
					/* Toggle up and down clipping */
					PPU_UpDown_Clip = ( PPU_UpDown_Clip ? 0 : 1 );
					break;

				case 'q':
				case 'Q':
					dwKeySystem |= PAD_SYS_QUIT;
					bRunning = false;
					break;

				case 'r':
				case 'R':
					/* Reset the application */
					Reset(); 
					break;

				case SDLK_PAGEUP:
					/* Increase Frame Skip */
					FrameSkip++;
					break;

				case SDLK_PAGEDOWN:
					/* Decrease Frame Skip */
					if ( FrameSkip > 0 )
					{
						FrameSkip--;
					}
					break;

				case 'm':
				case 'M':
					/* Toggle of sound mute */
					APU_Mute = ( APU_Mute ? 0 : 1 );
					break;

				case 'i':
				case 'I':
					/* If emulation thread doesn't run, nothing here */
					if ( bRunning )
					{
						InfoNES_MessageBox( "Mapper : %d\nPRG ROM : %dKB\nCHR ROM : %dKB\n" \
							"Mirroring : %s\nSRAM : %s\n4 Screen : %s\nTrainer : %s\n",
							MapperNo, NesHeader.byRomSize * 16, NesHeader.byVRomSize * 8,
							( ROM_Mirroring ? "V" : "H" ), ( ROM_SRAM ? "Yes" : "No" ), 
							( ROM_FourScr ? "Yes" : "No" ), ( ROM_Trainer ? "Yes" : "No" ) );
					}
					break;

				case 'v':
				case 'V':
					/* Version Infomation */
					InfoNES_MessageBox( "%s\nA fast and portable NES emulator\n"
						"Copyright (c) 1999-2005 Jay's Factory <jays_factory@excite.co.jp>",
						VERSION );
					break;

				default:  
					break;
			}
			break;
		case SDL_KEYUP:
			switch ( event.key.keysym.sym ) {
				case SDLK_RIGHT:
					dwKeyPad1 &= ~( 1 << 7 );
					break;

				case SDLK_LEFT:
					dwKeyPad1 &= ~( 1 << 6 );
					break;

				case SDLK_DOWN:
					dwKeyPad1 &= ~( 1 << 5 );
					break;
					
				case SDLK_UP:
					dwKeyPad1 &= ~( 1 << 4 );
					break;

				case 's':
				case 'S':
					/* Start */
					dwKeyPad1 &= ~( 1 << 3 );
					break;

				case 'a':
				case 'A':
					/* Select */
					dwKeyPad1 &= ~( 1 << 2 );
					break;

				case 'z':
				case 'Z': 
					/* 'A' */
					dwKeyPad1 &= ~( 1 << 1 );
					break;

				case 'x': 
				case 'X': 
					/* 'B' */
					dwKeyPad1 &= ~( 1 << 0 );
					break;

				default:  
					break;
			}
			break;
#ifdef GP2X
		case SDL_JOYBUTTONDOWN:
			joydown[event.jbutton.button] = true;
			if(joydown[0x0A] && joydown[0x0B] && joydown[0x12]) {
				/* Toggle up and down clipping */
				PPU_UpDown_Clip = ( PPU_UpDown_Clip ? 0 : 1 );
				joydown[0x12] = 0;
			}
			if(joydown[0x0B] && joydown[0x12]) {
				dwKeySystem |= PAD_SYS_QUIT;
				bRunning = false;
				memset(joydown, 0, sizeof(joydown));
			}
			if(joydown[0x0B] && joydown[0x08]) {
				Reset();
				joydown[0x08] = 0;
			}
			if(joydown[0x0B] && joydown[0x09]) {
				/* Toggle of sound mute */
				APU_Mute = ( APU_Mute ? 0 : 1 );
				joydown[0x09] = 0;
			}
			if(joydown[0x0B] && joydown[0x10]) {
				FrameSkip++;
				joydown[0x10] = 0;
			} else if(joydown[0x10]){
				main_volume += 15;
				if(main_volume > 128) main_volume = 128;
				joydown[0x10] = 0;
			}
			if(joydown[0x0B] && joydown[0x11]) {
				if ( FrameSkip > 0 )
					FrameSkip--;
				joydown[0x11] = 0;
			} else if(joydown[0x11]) {
				main_volume -= 15;
				if(main_volume < 0) main_volume = 0;
				joydown[0x11] = 0;
			}
			Joystick();
			break;
		case SDL_JOYBUTTONUP:
			joydown[event.jbutton.button] = false;
			Joystick();
			break;
#endif
		}
	}
}

/*===================================================================*/
/*                                                                   */
/*     Finalize() : When invoked via signal delete_event             */
/*                                                                   */
/*===================================================================*/
void Finalize( void )
{
	dwKeySystem |= PAD_SYS_QUIT;
	/* Save SRAM*/
	SaveSRAM();
	SDL_Quit();
#ifdef GP2X
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif
}

/*===================================================================*/
/*                                                                   */
/*     Reset() : Reset NES Hardware                                  */
/*                                                                   */
/*===================================================================*/
void Reset( void )
{
	if(bRunning) {
		dwKeySystem |= PAD_SYS_QUIT; 

		/* Save SRAM File */
		SaveSRAM();

		/* Load cassette */
		if ( InfoNES_Load ( szRomName ) == 0 )
		{ 
			/* Load SRAM */
			LoadSRAM();
		}
	}
}

/*===================================================================*/
/*                                                                   */
/*           LoadSRAM() : Load a SRAM                                */
/*                                                                   */
/*===================================================================*/
int LoadSRAM()
{
/*
 *  Load a SRAM
 *
 *  Return values
 *     0 : Normally
 *    -1 : SRAM data couldn't be read
 */

	FILE *fp;
	unsigned char pSrcBuf[ SRAM_SIZE ];
	unsigned char chData;
	unsigned char chTag;
	int nRunLen;
	int nDecoded;
	int nDecLen;
	int nIdx;

	// It doesn't need to save it
	nSRAM_SaveFlag = 0;

	// It is finished if the ROM doesn't have SRAM
	if ( !ROM_SRAM )
		return 0;

	// There is necessity to save it
	nSRAM_SaveFlag = 1;

	// The preparation of the SRAM file name
	strcpy( szSaveName, szRomName );
	strcpy( strrchr( szSaveName, '.' ) + 1, "srm" );

	/*-------------------------------------------------------------------*/
	/*  Read a SRAM data                                                 */
	/*-------------------------------------------------------------------*/

	// Open SRAM file
	fp = fopen( szSaveName, "rb" );
	if ( fp == NULL )
		return -1;

	// Read SRAM data
	fread( pSrcBuf, SRAM_SIZE, 1, fp );

	// Close SRAM file
	fclose( fp );

	/*-------------------------------------------------------------------*/
	/*  Extract a SRAM data                                              */
	/*-------------------------------------------------------------------*/

	nDecoded = 0;
	nDecLen = 0;

	chTag = pSrcBuf[ nDecoded++ ];

	while ( nDecLen < 8192 )
	{
		chData = pSrcBuf[ nDecoded++ ];

		if ( chData == chTag )
		{
			chData = pSrcBuf[ nDecoded++ ];
			nRunLen = pSrcBuf[ nDecoded++ ];
			for ( nIdx = 0; nIdx < nRunLen + 1; ++nIdx )
			{
				SRAM[ nDecLen++ ] = chData;
			}
		}
		else
		{
			SRAM[ nDecLen++ ] = chData;
		}
	}

	// Successful
	return 0;
}

/*===================================================================*/
/*                                                                   */
/*           SaveSRAM() : Save a SRAM                                */
/*                                                                   */
/*===================================================================*/
int SaveSRAM()
{
/*
 *  Save a SRAM
 *
 *  Return values
 *     0 : Normally
 *    -1 : SRAM data couldn't be written
 */

	FILE *fp;
	int nUsedTable[ 256 ];
	unsigned char chData;
	unsigned char chPrevData;
	unsigned char chTag;
	int nIdx;
	int nEncoded;
	int nEncLen;
	int nRunLen;
	unsigned char pDstBuf[ SRAM_SIZE ];

	if ( !nSRAM_SaveFlag )
		return 0;  // It doesn't need to save it

	/*-------------------------------------------------------------------*/
	/*  Compress a SRAM data                                             */
	/*-------------------------------------------------------------------*/

	memset( nUsedTable, 0, sizeof nUsedTable );

	for ( nIdx = 0; nIdx < SRAM_SIZE; ++nIdx )
	{
		++nUsedTable[ SRAM[ nIdx++ ] ];
	}
	for ( nIdx = 1, chTag = 0; nIdx < 256; ++nIdx )
	{
		if ( nUsedTable[ nIdx ] < nUsedTable[ chTag ] )
			chTag = nIdx;
	}

	nEncoded = 0;
	nEncLen = 0;
	nRunLen = 1;

	pDstBuf[ nEncLen++ ] = chTag;

	chPrevData = SRAM[ nEncoded++ ];

	while ( nEncoded < SRAM_SIZE && nEncLen < SRAM_SIZE - 133 )
	{
		chData = SRAM[ nEncoded++ ];

		if ( chPrevData == chData && nRunLen < 256 )
			++nRunLen;
		else
		{
			if ( nRunLen >= 4 || chPrevData == chTag )
			{
				pDstBuf[ nEncLen++ ] = chTag;
				pDstBuf[ nEncLen++ ] = chPrevData;
				pDstBuf[ nEncLen++ ] = nRunLen - 1;
			}
			else
			{
				for ( nIdx = 0; nIdx < nRunLen; ++nIdx )
					pDstBuf[ nEncLen++ ] = chPrevData;
			}

			chPrevData = chData;
			nRunLen = 1;
		}

	}
	if ( nRunLen >= 4 || chPrevData == chTag )
	{
		pDstBuf[ nEncLen++ ] = chTag;
		pDstBuf[ nEncLen++ ] = chPrevData;
		pDstBuf[ nEncLen++ ] = nRunLen - 1;
	}
	else
	{
		for ( nIdx = 0; nIdx < nRunLen; ++nIdx )
			pDstBuf[ nEncLen++ ] = chPrevData;
	}

	/*-------------------------------------------------------------------*/
	/*  Write a SRAM data                                                */
	/*-------------------------------------------------------------------*/

	// Open SRAM file
	fp = fopen( szSaveName, "wb" );
	if ( fp == NULL )
		return -1;

	// Write SRAM data
	fwrite( pDstBuf, nEncLen, 1, fp );

	// Close SRAM file
	fclose( fp );

	// Successful
	return 0;
}

/*===================================================================*/
/*                                                                   */
/*                  InfoNES_Menu() : Menu screen                     */
/*                                                                   */
/*===================================================================*/
int InfoNES_Menu()
{
/*
 *  Menu screen
 *
 *  Return values
 *     0 : Normally
 *    -1 : Exit InfoNES
 */

	/* If terminated */
	if ( !bRunning )
	{
		return -1;
	}

	/* Nothing to do here */
	return 0;
}

/*===================================================================*/
/*                                                                   */
/*               InfoNES_ReadRom() : Read ROM image file             */
/*                                                                   */
/*===================================================================*/
int InfoNES_ReadRom( const char *pszFileName )
{
/*
 *  Read ROM image file
 *
 *  Parameters
 *    const char *pszFileName          (Read)
 *
 *  Return values
 *     0 : Normally
 *    -1 : Error
 */

	FILE *fp;

	/* Open ROM file */
	fp = fopen( pszFileName, "rb" );
	if ( fp == NULL )
		return -1;

	/* Read ROM Header */
	fread( &NesHeader, sizeof NesHeader, 1, fp );
	if ( memcmp( NesHeader.byID, "NES\x1a", 4 ) != 0 )
	{
		/* not .nes file */
		fclose( fp );
		return -1;
	}

	/* Clear SRAM */
	memset( SRAM, 0, SRAM_SIZE );

	/* If trainer presents Read Triner at 0x7000-0x71ff */
	if ( NesHeader.byInfo1 & 4 )
	{
		fread( &SRAM[ 0x1000 ], 512, 1, fp );
	}

	/* Allocate Memory for ROM Image */
	ROM = (BYTE *)malloc( NesHeader.byRomSize * 0x4000 );

	/* Read ROM Image */
	fread( ROM, 0x4000, NesHeader.byRomSize, fp );

	if ( NesHeader.byVRomSize > 0 )
	{
		/* Allocate Memory for VROM Image */
		VROM = (BYTE *)malloc( NesHeader.byVRomSize * 0x2000 );

		/* Read VROM Image */
		fread( VROM, 0x2000, NesHeader.byVRomSize, fp );
	}

	/* File close */
	fclose( fp );

	/* Successful */
	return 0;
}

/*===================================================================*/
/*                                                                   */
/*           InfoNES_ReleaseRom() : Release a memory for ROM         */
/*                                                                   */
/*===================================================================*/
void InfoNES_ReleaseRom()
{
/*
 *  Release a memory for ROM
 *
 */

	if ( ROM )
	{
		free( ROM );
		ROM = NULL;
	}

	if ( VROM )
	{
		free( VROM );
		VROM = NULL;
	}
}

/*===================================================================*/
/*                                                                   */
/*             InfoNES_MemoryCopy() : memcpy                         */
/*                                                                   */
/*===================================================================*/
void *InfoNES_MemoryCopy( void *dest, const void *src, int count )
{
/*
 *  memcpy
 *
 *  Parameters
 *    void *dest                       (Write)
 *      Points to the starting address of the copied block's destination
 *
 *    const void *src                  (Read)
 *      Points to the starting address of the block of memory to copy
 *
 *    int count                        (Read)
 *      Specifies the size, in bytes, of the block of memory to copy
 *
 *  Return values
 *    Pointer of destination
 */

	memcpy( dest, src, count );
	return dest;
}

/*===================================================================*/
/*                                                                   */
/*             InfoNES_MemorySet() : memset                          */
/*                                                                   */
/*===================================================================*/
void *InfoNES_MemorySet( void *dest, int c, int count )
{
/*
 *  memset
 *
 *  Parameters
 *    void *dest                       (Write)
 *      Points to the starting address of the block of memory to fill
 *
 *    int c                            (Read)
 *      Specifies the byte value with which to fill the memory block
 *
 *    int count                        (Read)
 *      Specifies the size, in bytes, of the block of memory to fill
 *
 *  Return values
 *    Pointer of destination
 */

	memset( dest, c, count);  
	return dest;
}

/*===================================================================*/
/*                                                                   */
/*      InfoNES_LoadFrame() :                                        */
/*           Transfer the contents of work frame on the screen       */
/*                                                                   */
/*===================================================================*/
void InfoNES_LoadFrame()
{
/*
 *  Transfer the contents of work frame on the screen
 *
 */
	Uint16 *pDest, *pSrc;
	int x, y;

	SDL_LockSurface(screen);
	pDest = (Uint16 *)screen->pixels + ((320 - NES_DISP_WIDTH) >> 1);
	pSrc = WorkFrame;

	/* Blit screen */
	for ( y = 0; y < NES_DISP_HEIGHT; y++ )
	{
		for ( x = 0; x < NES_DISP_WIDTH; x++ )
		{  
			*pDest = (((*pSrc) & 0x7FE0) << 1) | ((*pSrc) & 0x001F);
			pDest++;pSrc++;
		}
		pDest += 320 - NES_DISP_WIDTH;
	}
	SDL_UnlockSurface(screen);
	SDL_Flip(screen);
}

/*===================================================================*/
/*                                                                   */
/*             InfoNES_PadState() : Get a joypad state               */
/*                                                                   */
/*===================================================================*/
void InfoNES_PadState( DWORD *pdwPad1, DWORD *pdwPad2, DWORD *pdwSystem )
{
/*
 *  Get a joypad state
 *
 *  Parameters
 *    DWORD *pdwPad1                   (Write)
 *      Joypad 1 State
 *
 *    DWORD *pdwPad2                   (Write)
 *      Joypad 2 State
 *
 *    DWORD *pdwSystem                 (Write)
 *      Input for InfoNES
 *
 */

	/* Transfer joypad state */
	*pdwPad1   = dwKeyPad1;
	*pdwPad2   = dwKeyPad2;
	*pdwSystem = dwKeySystem;
}

/*===================================================================*/
/*                                                                   */
/*        InfoNES_SoundInit() : Sound Emulation Initialize           */
/*                                                                   */
/*===================================================================*/
void InfoNES_SoundInit( void ) 
{
	SDL_PauseAudio(1);
	sound_on = false;
}

/*===================================================================*/
/*                                                                   */
/*        InfoNES_SoundOpen() : Sound Open                           */
/*                                                                   */
/*===================================================================*/
void InfoNES_SoundCallback(void *data, Uint8 *buff, int len)
{
	if( sound_on && main_volume ) {
		SDL_LockAudio();
		if (main_volume < 128) {
		  apu_process(final_wave, len >> 1);
		  SDL_MixAudio (buff, final_wave, len, main_volume);
		} else {
		  apu_process(buff, len >> 1);
		}
		SDL_UnlockAudio();
	}
	sync_apu_register();
/*	if ( sound_on && wavflag && main_volume )
	{
		SDL_LockAudio();
		SDL_MixAudio (buff, &final_wave[(wavflag - 1) << 10], 1024, main_volume);
		wavflag = 0;
		SDL_UnlockAudio();
	}*/
}

int InfoNES_SoundOpen( int samples_per_sync, int sample_rate ) 
{
	SDL_AudioSpec spec;

	if(!sound_open) {
		spec.format = AUDIO_S16;
		spec.channels = 1;
		spec.freq = sample_rate;
		spec.samples = 1024;

		spec.callback = InfoNES_SoundCallback;
		spec.userdata = NULL;

		if ((!sound_open) && (SDL_OpenAudio(&spec, NULL) < 0)) {
			printf("Unable to open audio: %s.\n", SDL_GetError());
			sound_on = false;
			return 0;
		}
/*
		printf ("Sound Open: Rate: %d, Buffer size: %d, 16-bit: %s, Stereo: %s\n",
			spec.freq, spec.samples, (spec.format == AUDIO_U8) ? "no": "yes",
			(spec.channels == 2) ? "yes" : "no");
*/
		sound_open = true;
	}

	SDL_PauseAudio(0);

	sound_on = true;

	return 1;
}

/*===================================================================*/
/*                                                                   */
/*        InfoNES_SoundClose() : Sound Close                         */
/*                                                                   */
/*===================================================================*/
void InfoNES_SoundClose( void ) 
{
	SDL_PauseAudio(1);
	sound_on = false;
}

/*===================================================================*/
/*                                                                   */
/*            InfoNES_SoundOutput() : Sound Output 5 Waves           */           
/*                                                                   */
/*===================================================================*/
void InfoNES_SoundOutput( int samples, BYTE *wave1, BYTE *wave2, BYTE *wave3, BYTE *wave4, BYTE *wave5 )
{
	int i;

	if ( sound_on ) 
	{
		for (i = 0; i < samples; i++) 
		{
#if 1
			final_wave[ waveptr ] = 
				( wave1[i] + wave2[i] + wave3[i] + wave4[i] + wave5[i] ) / 5;
#else
			final_wave[ waveptr ] = wave4[i];
#endif


			waveptr++;
			if ( waveptr == 2048 ) 
			{
				waveptr = 0;
				wavflag = 2;
			} 
			else if ( waveptr == 1024 )
			{
				wavflag = 1;
			}
		}
	}
}

/*===================================================================*/
/*                                                                   */
/*            InfoNES_Wait() : Wait Emulation if required            */
/*                                                                   */
/*===================================================================*/
void InfoNES_Wait()
{
	SDL_ProcessEvent();
#ifdef WIN32
	Sleep(0);
#endif
}

/*===================================================================*/
/*                                                                   */
/*            InfoNES_MessageBox() : Print System Message            */
/*                                                                   */
/*===================================================================*/
void InfoNES_MessageBox( char *pszMsg, ... )
{
	char pszErr[ 1024 ];
	va_list args;

	// Create the message body
	va_start( args, pszMsg );
	vsprintf( pszErr, pszMsg, args );  pszErr[ 1023 ] = '\0';
	va_end( args );

	// Only for debug
	printf( pszErr );
}

/*
 * End of InfoNES_System_SDL.cpp
 */
