#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef WIN32
	#include "win32/dirent.h"
	#define strcasecmp stricmp
#else
	#include <unistd.h>
	#include <dirent.h>
#endif

#include <SDL.h>
#include <SDL_image.h>

#include "filer.h"
//#include "port.h"

extern "C" {
#ifdef WIN32
	#include "UnicodeFont.h"
#else
	#include <UnicodeFont.h>
#endif
}

#ifdef GP2X
#ifndef __GP2X__
#define __GP2X__
#endif
#endif

#ifndef S_ISDIR
#define S_ISDIR(m) (((m) & _S_IFDIR) == _S_IFDIR)
#endif

#define MAX_NAME 256
#define MAX_ENTRY 512
#define LIST_COLORKEY 0xFEFE

extern SDL_Surface *screen;
extern int main_volume;

SDL_Surface *back = NULL;
SDL_Surface *list = NULL;

SDL_Surface *icon_game = NULL;
SDL_Surface *icon_zipped = NULL;
SDL_Surface *icon_folder = NULL;

int skin_no;
int use_icon;
int draw_pos, cur_pos, redraw;
static char open_path[_MAX_PATH];
static char skin_path[_MAX_PATH];
static char work_path[_MAX_PATH];

struct filelist {
	char name[MAX_NAME + 1];
	int isdir;
} files[MAX_ENTRY];
static int nfiles;

enum {    
	EXT_NES,
	EXT_ZIPPED,
	EXT_UNKNOWN  
};

const struct {
	char *szExt;
	int nExtId;
} stExtentions[] = {        
    "nes", EXT_NES,
	NULL, EXT_UNKNOWN
};

int getExtId(const char *szFilePath) {
	char *pszExt;
	int i;
	if((pszExt = strrchr(szFilePath, '.'))) {
		pszExt++;				
		for (i = 0; stExtentions[i].nExtId != EXT_UNKNOWN; i++) {
			if (!strcasecmp(stExtentions[i].szExt,pszExt)) {
				return stExtentions[i].nExtId;
			}
		}
	}
	return EXT_UNKNOWN;
}

int cmpFile(struct filelist *a, struct filelist *b)
{
	if(a->isdir == b->isdir){
		return strcasecmp(a->name, b->name);
	} else if(a->isdir)
		return -1;
	else
		return 1;
}

void sort(int left, int right) {
	struct filelist tmp, pivot;
	int i, p;
	
	if (left < right) {
		pivot = files[left];
		p = left;
		for (i=left+1; i<=right; i++) {
			if (cmpFile(&files[i],&pivot)<0){
				p=p+1;
				tmp=files[p];
				files[p]=files[i];
				files[i]=tmp;
			}
		}
		files[left] = files[p];
		files[p] = pivot;
		sort(left, p-1);
		sort(p+1, right);
	}
}

void getDir(const char *path, int updir) {
	strcpy(open_path, path);
	
	while(1) {
		DIR *fd;
		char prev_dir[MAX_NAME];
		int prev_len = 0;

		if(updir && (strcasecmp(open_path, "/mnt/") == 0) ) {
			return;
		}
		if(updir) {
			int i;
			for(i = strlen(open_path)-1; i>0 ; i--) {
				if((open_path[i-1] == '/') || (open_path[i-1] == '\\')) {
					strcpy(prev_dir, &open_path[i]);
					prev_len = strlen(prev_dir);
					while( (prev_len > 0) &&
						((prev_dir[prev_len-1] == '/') || (prev_dir[prev_len-1] == '\\')) ) {
						prev_dir[prev_len-1] = '\0';
						--prev_len;
					}
					if(prev_len != 0) {
						prev_dir[prev_len++] = '/';
						prev_dir[prev_len] = '\0';
						printf("Filer: Prev Dir: %s\n", prev_dir);
					}
					open_path[i] = '\0';
					updir = 0;
					break;
				}
			}
			if(updir){
				return;
			}
		}

		fd = opendir(open_path);
		if(fd != NULL) {
			struct dirent *direntp;
			struct stat statbuf;
			char fullname[_MAX_PATH];

			nfiles = 0;
			while( (nfiles < MAX_ENTRY) && (direntp = readdir(fd)) != NULL ){
				if(direntp->d_name[0] == '.') continue;
				sprintf(fullname, "%s%s", open_path, direntp->d_name);
				if(stat(fullname, &statbuf) == 0) {
					files[nfiles].isdir = S_ISDIR(statbuf.st_mode);
					if(files[nfiles].isdir) {
						strcpy(files[nfiles].name, direntp->d_name);
						strcat(files[nfiles].name, "/");
						nfiles++;
					} else if(getExtId(direntp->d_name) != EXT_UNKNOWN) {
						strcpy(files[nfiles].name, direntp->d_name);
						nfiles++;
					}
				}
			}
			closedir(fd);

			sort(0, nfiles-1);

			cur_pos = 0;
			draw_pos = 0;
			redraw = 1;
			if(prev_len != 0){
				int i;
				for(i=0;i<nfiles;i++){
					if(strcasecmp(files[i].name, prev_dir) == 0) {
						cur_pos = i;
						draw_pos = i - 2;
						if((draw_pos + 5) > nfiles) draw_pos = nfiles - 5;
						if(draw_pos < 0) draw_pos = 0;
						break;
					}
				}
			}
			printf("Filer: Get Dir %s(%d).\n", open_path, nfiles);

			return;
		}
		updir = 1;
	}
}

void print(SDL_Surface *surface, int x, int y, Uint16 *str, int col) {
	unsigned short *pixel = (unsigned short *)surface->pixels;
	unsigned short font[16];
	Uint16 code;
	int h, w;
	int j, k;
	int ptr;
	int pitch = surface->pitch >> 1;
	int i = 0;
	while(str[i]) {
		code = str[i++];
		GetFont(code, font, &h, &w);
		if(y>200) {
			if((x+w) >= 320) {
				x = 0;
				y += h;
			}
		} else {
			if((x+w) >= 300) {
				break;
			}
		}
		ptr = (pitch * y) + x;
		for(j = 0;j < h;j++) {
			if(font[j] != 0) {
				for(k = 0;k < w;k++) {
					if(font[j] & (0x8000 >> k))
						*(pixel + ptr + k) = col;
				}
			}
			ptr += pitch;
		}
		x += w;
	}
}

Uint16 *U8toU16(Uint16 *unicode, Uint8 *utf8, int len)
{
	int i, j;
	Uint16 ch;

	for ( i=0, j=0; i < len; ++i, ++j ) {
		ch = ((const unsigned char *)utf8)[i];
		if ( ch >= 0xF0 ) {
			ch  =  (Uint16)(utf8[i]&0x07) << 18;
			ch |=  (Uint16)(utf8[++i]&0x3F) << 12;
			ch |=  (Uint16)(utf8[++i]&0x3F) << 6;
			ch |=  (Uint16)(utf8[++i]&0x3F);
		} else
		if ( ch >= 0xE0 ) {
			ch  =  (Uint16)(utf8[i]&0x0F) << 12;
			ch |=  (Uint16)(utf8[++i]&0x3F) << 6;
			ch |=  (Uint16)(utf8[++i]&0x3F);
		} else
		if ( ch >= 0xC0 ) {
			ch  =  (Uint16)(utf8[i]&0x1F) << 6;
			ch |=  (Uint16)(utf8[++i]&0x3F);
		}
		unicode[j] = ch;
	}
	unicode[j] = 0;

	return unicode;
}

void printU8(SDL_Surface *surface, int x, int y, char *u8, int col) {
	int len = strlen(u8);
	Uint16 str[_MAX_PATH];
	U8toU16(str, (Uint8 *)u8, len);
	print(surface, x, y, str, col);
}

void drawlist() {
	int y = 70;
	int i;
	SDL_Rect dstrect;
	dstrect.x = 20;

	if(nfiles > 5) {
		if((draw_pos + 4) < cur_pos)
			draw_pos = cur_pos - 4;
		if(draw_pos > cur_pos)
			draw_pos = cur_pos;
	}
	
	SDL_FillRect(list, NULL, LIST_COLORKEY);
	for(i=draw_pos;(i<nfiles) && (y < 200);i++, y+=30){
		SDL_LockSurface(list);
		if(i == cur_pos)
			printU8(list, 50, y + 8, files[i].name, 0x0000);
		else
			printU8(list, 50, y + 8, files[i].name, 0xFFFF);
		SDL_UnlockSurface(list);
		dstrect.y = y;
		if(files[i].isdir)
			SDL_BlitSurface(icon_folder, NULL, list, &dstrect);
		else
			switch(getExtId(files[i].name)) {
			case EXT_NES:
				SDL_BlitSurface(icon_game, NULL, list, &dstrect);
				break;
			case EXT_ZIPPED:
				SDL_BlitSurface(icon_zipped, NULL, list, &dstrect);
				break;
			}
	}

	SDL_BlitSurface(back, NULL, screen, NULL);
	SDL_BlitSurface(list, NULL, screen, NULL);
	SDL_Flip(screen);
}

void defaultskin() {
	strcpy(skin_path, work_path);

	if(icon_game) SDL_FreeSurface(icon_game);
	if(icon_zipped) SDL_FreeSurface(icon_zipped);
	if(icon_folder) SDL_FreeSurface(icon_folder);
	icon_game = icon_zipped = icon_folder = NULL;

	use_icon = 0;

	skin_no = 0;
}

int skinload(int no) {
	char file[100];
	struct stat statbuf;
	SDL_Surface *bg;

	sprintf(skin_path, "%sskin%d", work_path, no);
	if(stat(skin_path, &statbuf) != 0) return 0;
	if(!S_ISDIR(statbuf.st_mode)) return 0;
	strcat(skin_path, "/");
	if(skin_no == no) return 1;
	skin_no = no;

	strcpy(file, skin_path);strcat(file, "filer_nes.png");
	bg = IMG_Load(file);
	if(!bg) {
		strcpy(file, skin_path);strcat(file, "filer.png");
		bg = IMG_Load(file);
	}
	if (bg) {
		SDL_BlitSurface(bg, NULL, back, NULL);
		SDL_FreeSurface(bg);
	}

	SDL_LockSurface(back);
	printU8(back, 188, 226, "Compile : " __DATE__, 0xFFFF);
	SDL_UnlockSurface(back);

	SDL_SetColorKey(list, SDL_SRCCOLORKEY, LIST_COLORKEY);

	if(icon_game) SDL_FreeSurface(icon_game);
	strcpy(file, skin_path);strcat(file, "game.png");
	icon_game = IMG_Load(file);

	if(icon_zipped) SDL_FreeSurface(icon_zipped);
	strcpy(file, skin_path);strcat(file, "zipped.png");
	icon_zipped = IMG_Load(file);

	if(icon_folder) SDL_FreeSurface(icon_folder);
	strcpy(file, skin_path);strcat(file, "folder.png");
	icon_folder = IMG_Load(file);

	use_icon = icon_game || icon_zipped ||icon_folder;

	redraw = 1;

	return 1;
}

void prevskin() {
	int no = skin_no - 1;
	if(no < 0) no = 9;
	while(no > 0) {
		if(skinload(no))
			return;
		no--;
	}
	if(!skinload(0))
		defaultskin();
}

void nextskin() {
	int no = skin_no + 1;
	while(no < 10) {
		if(skinload(no))
			return;
		no++;
	}
	if(!skinload(0))
		defaultskin();
}

int eventloop() {
	SDL_Event event;
	int flag = 0;
	static int move_dir = 0;
	static Uint32 move_tick = 0;
	while(SDL_PollEvent(&event)) {
		switch(event.type) {
		case SDL_KEYDOWN:
			switch (event.key.keysym.sym) {
			case SDLK_UP:
				move_dir = -1;
				move_tick = 0;
				break;
			case SDLK_DOWN:
				move_dir = 1;
				move_tick = 0;
				break;
			case SDLK_LEFT:
				move_dir = 0;
				getDir(open_path, 1);
				break;
			case SDLK_RIGHT:
			case SDLK_RETURN:
				if(files[cur_pos].isdir){
					strcat(open_path, files[cur_pos].name);
					getDir(open_path, 0);
					move_dir = 0;
					cur_pos = 0;
					draw_pos = 0;
					redraw = 1;
				} else {
					flag = 1;
				}
				break;
			case SDLK_ESCAPE:
			case 'q':
				flag = 2;
				break;
			case 'z':
				prevskin();
				break;
			case 'x':
				nextskin();
				break;
			}
			break;
		case SDL_KEYUP:
			switch (event.key.keysym.sym) {
			case SDLK_UP:
			case SDLK_DOWN:
				move_dir = 0;
				move_tick = 0;
				break;
			}
			break;
#ifdef __GP2X__
		case SDL_JOYBUTTONDOWN:
			switch(event.jbutton.button) {
			case 0x00: // Up
				move_dir = -1;
				move_tick = 0;
				break;
			case 0x04: // Down
				move_dir = 1;
				move_tick = 0;
				break;
			case 0x02: // Left
			case 0x0E: // X Button
				move_dir = 0;
				getDir(open_path, 1);
				break;
			case 0x06: // Right
			case 0x0D: // B Button
			case 0x12: // Stick
				if(files[cur_pos].isdir){
					strcat(open_path, files[cur_pos].name);
					getDir(open_path, 0);
					move_dir = 0;
					cur_pos = 0;
					draw_pos = 0;
					redraw = 1;
				} else {
					flag = 1;
				}
				break;
			case 0x08: // Start Button
				flag = 2;
				break;
			case 0x0A: // L Button
				prevskin();
				break;
			case 0x0B: // R Button
				nextskin();
				break;
			}
			if(event.jbutton.button == 0x10) {
				main_volume += 15;
				if(main_volume > 128) main_volume = 128;
			} else if(event.jbutton.button == 0x11) {
				main_volume -= 15;
				if(main_volume < 0) main_volume = 0;
			}
			break;
		case SDL_JOYBUTTONUP:
			switch(event.jbutton.button) {
			case 0x00: // Up
			case 0x04: // Down
				move_dir = 0;
				move_tick = 0;
				break;
			}
			break;
#endif
		}
	}

	if( (move_dir != 0) && ((move_tick + 70) < SDL_GetTicks()) ) {
		static int boost = 0;
		static Uint32 boost_tick = 0;
		int old_cur_pos = cur_pos;

		if( (boost) || ((move_tick + 200) < SDL_GetTicks()) ) {
			cur_pos += move_dir;
			if(cur_pos >= nfiles) cur_pos =  nfiles - 1;
			if(cur_pos < 0) cur_pos = 0;
 
			redraw = (old_cur_pos != cur_pos);

			if(move_tick == 0) {
				boost_tick = SDL_GetTicks();
				boost = 0;
			}
			move_tick = SDL_GetTicks();
		}

		if( (!boost) && ((boost_tick + 300) < SDL_GetTicks()) ) {
			boost = 1;
		}
	}

	return flag;
}

int filer(const char *cwd, const char *path, char *out, int *skin) {
	int flag;
	back = SDL_CreateRGBSurface(0, 320, 240, 16,
		0xF800, 0x07E0,	0x001F, 0x0000);
	list = SDL_CreateRGBSurface(0, 320, 240, 16,
		0xF800, 0x07E0,	0x001F, 0x0000);

	strcpy(work_path, cwd);
	skin_no = -1;
	if( (!skinload(*skin)) && (!skinload(0)) ) {
		defaultskin();
	}

	getDir(path, 0);
	cur_pos = 0;
	draw_pos = 0;
	redraw = 1;
	while((flag = eventloop()) == 0) {
		if(redraw) {
			drawlist();
			redraw = 0;
		}
	}

	SDL_FreeSurface(icon_folder);
	SDL_FreeSurface(icon_zipped);
	SDL_FreeSurface(icon_game);
	SDL_FreeSurface(list);
	SDL_FreeSurface(back);
	back = icon_game = icon_zipped = icon_folder = NULL;

	if(flag == 1) {
		// Loading Image...
		SDL_Surface *loading;
		sprintf(skin_path, "%sskin%d/loading.png", work_path, skin_no);
		loading = IMG_Load(skin_path);
		if(loading) {
			if( (loading->w < 320) || (loading->h < 240) ) {
				SDL_Rect dstrect;
				dstrect.w = loading->w;
				dstrect.h = loading->h;
				if(dstrect.w < 320)
					dstrect.x = (320 - dstrect.w) >> 1;
				else
					dstrect.x = 0;
				if(dstrect.h < 240)
					dstrect.y = (240 - dstrect.h) >> 1;
				else
					dstrect.y = 0;
				SDL_BlitSurface(loading, NULL, screen, &dstrect);
			} else {
				SDL_BlitSurface(loading, NULL, screen, NULL);
			}
			SDL_FreeSurface(loading);
			SDL_Flip(screen);
		}

		strcpy(out, open_path);
		strcat(out, files[cur_pos].name);
		*skin = skin_no;

		return 1;
	}

	out[0] = '\0';
	return 0;
}
