/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
 * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
 *
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: http://www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */
#ifdef WIN32
	#include <windows.h>
#endif
#include <signal.h>
#include <errno.h>

#include <stdlib.h>
#include <string.h>
#ifndef WIN32
	#include <unistd.h>
	#include <sys/time.h>
	#include <sys/ioctl.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <ctype.h>
#ifdef WIN32
	#include "win32/dirent.h"
#else
	#include <dirent.h>
	#include <sys/soundcard.h>
#endif
#include <SDL.h>
#include <SDL_image.h>
#include <SDL_inifile.h>
#include "keydef.h"

#include "snes9x.h"
#include "port.h"
#include "memmap.h"
#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu.h"
#include "display.h"
#include "gfx.h"
#include "soundux.h"
#include "spc700.h"

#include "filer.h"

#define bool8_32 bool8

uint8 *keyssnes;
bool8 joydown[32];
int OldSkipFrame;
bool8 sdl_init = 0;
int main_volume;

static bool8 running;
static char work_path[_MAX_PATH];
static char save_path[_MAX_PATH];
static char roms_path[_MAX_PATH];
static char config_file[_MAX_PATH];

void InitTimer ();
void *S9xProcessSound (void *);

extern void S9xDisplayFrameRate (uint8 *, uint32);
extern void S9xDisplayString (const char *string, uint8 *, uint32);
extern SDL_Surface *screen;

static uint32 ffc = 0;
bool nso = FALSE, vga = FALSE;
uint32 xs = 320, ys = 240, cl = 0, cs = 0, mfs = 10;
 

char rom_filename[_MAX_PATH];
char snapshot_filename[_MAX_PATH];
//char *snapshot_filename = NULL;

void OutOfMemory ()
{
	fprintf (stderr, "Snes9X: Memory allocation failure - not enough RAM/virtual memory available.\nS9xExiting...\n");
	Memory.Deinit ();
	S9xDeinitAPU ();

	exit (1);
}

void S9xParseArg (char **argv, int &i, int argc)
{

	if (strcasecmp (argv [i], "-b") == 0 ||
		strcasecmp (argv [i], "-bs") == 0 ||
		strcasecmp (argv [i], "-buffersize") == 0)
	{
	if (i + 1 < argc)
		Settings.SoundBufferSize = atoi (argv [++i]);
	else
		S9xUsage ();
	}
/*
	else if (strcmp (argv [i], "-l") == 0 ||
		strcasecmp (argv [i], "-loadsnapshot") == 0)
	{
	if (i + 1 < argc)
		snapshot_filename = argv [++i];
	else
		S9xUsage ();
	}
	else if (strcmp (argv [i], "-nso") == 0)
		nso = TRUE;
	else if (strcmp (argv [i], "-x2") == 0)
		vga = TRUE;
*/
	else if (strcmp (argv [i], "-xs") == 0)
	{
	if (i + 1 < argc)
		xs = atoi(argv [++i]);
	else
		S9xUsage ();
	}
	else if (strcmp (argv [i], "-ys") == 0)
	{
	if (i + 1 < argc)
		ys = atoi(argv [++i]);
	else
		S9xUsage ();
	}
	else if (strcmp (argv [i], "-cl") == 0)
	{
	if (i + 1 < argc)
		cl = atoi(argv [++i]);
	else
		S9xUsage ();
	}
	else if (strcmp (argv [i], "-cs") == 0)
	{
	if (i + 1 < argc)
		cs = atoi(argv [++i]);
	else
		S9xUsage ();
	}
	else if (strcmp (argv [i], "-mfs") == 0)
	{
	if (i + 1 < argc)
		mfs = atoi(argv [++i]);
	else
		S9xUsage ();
	}
	else
		S9xUsage ();
//	S9xParseDisplayArg (argv, i, argc);
}

void Finalize()
{
	if(sdl_init) {
		S9xDeinitDisplay ();
		INI_Close();
		SDL_Quit();
	}
#ifdef __GP2X__
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif
}

void SetVolume()
{
	int fd, v, devs;

	if ((fd = open("/dev/mixer", O_RDONLY)) > 0)
	{
		ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
		if (devs & (1 << SOUND_MIXER_PCM))
		{
			v = ((int)main_volume << 8) | (int)main_volume;
			ioctl(fd, MIXER_WRITE(SOUND_MIXER_PCM), &v);
		}
		close(fd);
	}
}

/*#include "cheats.h"*/
//extern "C"
int main (int argc, char **argv)
{
	int skin;
	FILE *fp;

	printf("snes9x for GP2X start...\n");
	atexit(Finalize);

#ifdef __unix__
	getcwd(work_path, _MAX_PATH);
	strcat(work_path, SLASH_STR);
#else
	int i;

	strcpy(work_path, argv[0]);
	for(i = strlen(work_path)-1; i>0 ; i--) {
		if((work_path[i-1] == '/') || (work_path[i-1] == '\\')) {
			work_path[i] = '\0';
			break;
		}
	}
#endif

	strcpy(save_path, work_path);
	strcat(save_path, "save");
	strcat(save_path, SLASH_STR);

	strcpy(roms_path, work_path);
	strcat(roms_path, "roms");
	strcat(roms_path, SLASH_STR);

	//strcpy(config_file, save_path);
	strcpy(config_file, ".");
	strcat(config_file, SLASH_STR);
	strcat(config_file, "snesgp2x.cfg");

	// main_volume = 80;

#ifdef __GP2X__
	system("/bin/mount -t vfat -o remount,sync,iocharset=utf8 /dev/mmcsd/disc0/part1 /mnt/sd");
#endif

	INI_Open(config_file);
	skin = INI_ReadInt("skin", "no", 0);
	main_volume = INI_ReadInt("sound", "volume", 80);
/*
	fp = fopen(config_file, "r");
	if(fp) {
		fscanf(fp, "%d", &skin);
		fclose(fp);
	} else
		skin = 0;
*/
	if((skin < 0) || (skin > 9)) skin = 0;
	if(main_volume < 0) main_volume = 0;
	else if(main_volume > 100) main_volume = 100;

	printf("Initialize...\n");
	S9xInitDisplay (argc, argv);
	S9xInitInputDevices ();

	while (1) {
		int flag;

		INI_WriteInt("sound", "volume", main_volume);

		// Initialize End... (Filer Start)
		if(!(flag = filer(work_path, roms_path, rom_filename, &skin))) {
			exit(0);
		} else {
			INI_WriteInt("skin", "no", skin);
			/*
			fp = fopen(config_file, "w");
			if(fp) {
				fprintf(fp, "%d\n", skin);
				fclose(fp);
			}
			*/
		}

		//os9x_asmcpu = (flag == 1);
		memset(joydown, 0, sizeof(joydown));

		printf("Select Rom: %s\n", rom_filename);

		// Emulator Start...

		ZeroMemory (&Settings, sizeof (Settings));

		Settings.JoystickEnabled = FALSE;
		Settings.SoundPlaybackRate = 4;
		Settings.Stereo = TRUE;
		Settings.SoundBufferSize = 1024;
		Settings.CyclesPercentage = 100;
		Settings.DisableSoundEcho = FALSE;
		Settings.APUEnabled = Settings.NextAPUEnabled = TRUE;
		Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
		Settings.SkipFrames = AUTO_FRAMERATE;
		Settings.ShutdownMaster = TRUE;
		Settings.FrameTimePAL = 20000;
		Settings.FrameTimeNTSC = 16667;
		Settings.FrameTime = Settings.FrameTimeNTSC;
		Settings.DisableSampleCaching = FALSE;
		Settings.DisableMasterVolume = FALSE;
		Settings.Mouse = FALSE;
		Settings.SuperScope = FALSE;
		Settings.MultiPlayer5 = FALSE;
	//	Settings.ControllerOption = SNES_MULTIPLAYER5;
		Settings.ControllerOption = 0;
		Settings.ForceTransparency = TRUE;
		Settings.Transparency = TRUE;
		Settings.SixteenBit = TRUE;
		Settings.SupportHiRes = FALSE;
		Settings.NetPlay = FALSE;
		Settings.ServerName [0] = 0;
		Settings.ThreadSound = TRUE;
		Settings.AutoSaveDelay = 30;
		Settings.ApplyCheats = TRUE;
		Settings.TurboMode = FALSE;
		Settings.TurboSkipFrames = 15;

		if (Settings.ForceNoTransparency)
			Settings.Transparency = FALSE;

		if (Settings.Transparency)
			Settings.SixteenBit = TRUE;

		Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

		if (!Memory.Init () || !S9xInitAPU())
			OutOfMemory ();

		uint32 saved_flags = CPU.Flags;

#ifdef GFX_MULTI_FORMAT
		S9xSetRenderPixelFormat (RGB565);
#endif
		// Temporary...
		Settings.APUEnabled = Settings.NextAPUEnabled = (flag == 1); // !os9x_asmcpu;

		if (!Settings.APUEnabled) {
			S9xSetSoundMute (TRUE);
		} else {
			SetVolume();
		}

		S9xInitSound (Settings.SoundPlaybackRate, Settings.Stereo,
			Settings.SoundBufferSize);

		printf("S9xGraphicsInit ();\n");
		if (!S9xGraphicsInit ())
			OutOfMemory ();

		// Loading...
		printf("Loading...\n");
		if (rom_filename)
		{
			if (!Memory.LoadROM (rom_filename))
			{
				char dir [_MAX_DIR + 1];
				char drive [_MAX_DRIVE + 1];
				char name [_MAX_FNAME + 1];
				char ext [_MAX_EXT + 1];
				char fname [_MAX_PATH + 1];

				_splitpath (rom_filename, drive, dir, name, ext);
				_makepath (fname, drive, dir, name, ext);

				strcpy (fname, S9xGetROMDirectory ());
				strcat (fname, SLASH_STR);
				strcat (fname, name);
				if (ext [0])
				{
					strcat (fname, ".");
					strcat (fname, ext);
				}
				_splitpath (fname, drive, dir, name, ext);
				_makepath (fname, drive, dir, name, ext);
				if (!Memory.LoadROM (fname))
				{
					printf ("Error opening: %s\n", rom_filename);
					exit (1);
				}
			}
			Memory.LoadSRAM (S9xGetFilename (".srm"));
//			S9xLoadCheatFile (S9xGetFilename (".cht"));
		}
		else
		{
			S9xReset ();
			Settings.Paused |= 2;
		}
		CPU.Flags = saved_flags;

		printf("Loading Snapshot...\n");
/*
		if (snapshot_filename)
		{
			int Flags = CPU.Flags & (DEBUG_MODE_FLAG | TRACE_FLAG);
			if (!S9xLoadSnapshot (snapshot_filename))
				exit (1);
			CPU.Flags |= Flags;
		}
*/
		strcpy(snapshot_filename, S9xGetFilename (".state"));

#ifndef __GP2X__
		S9xGraphicsMode ();
		sprintf (String, "\"%s\" %s: %s", Memory.ROMName, TITLE, VERSION);
		S9xSetTitle (String);
#endif

		printf ("\"%s\" %s: %s\n", Memory.ROMName, TITLE, VERSION);

		if (nso) {
			Settings.SoundBufferSize = 8192;
			Settings.SoundPlaybackRate = 1;
			Settings.DisableSoundEcho = TRUE;
			Settings.DisableMasterVolume = TRUE;
			Settings.Stereo = FALSE;
			S9xSetSoundMute (TRUE);
		} else if (!Settings.APUEnabled) {
			S9xSetSoundMute (FALSE);
		}
		Settings.SoundSync = FALSE;

		SDL_Surface *bg = IMG_Load("skin/gamebackground.bmp");
		if(bg) {
			SDL_BlitSurface(bg, NULL, screen, NULL);
			SDL_FreeSurface(bg);
			bg = NULL;
		} else {
			SDL_FillRect(screen, NULL, 0);
		}
		SDL_Flip(screen);
		printf("Running...\n");
		running = TRUE;
		while (running) {
#ifdef __GP2X__
			S9xMainLoop ();
#else
			if (!Settings.Paused) {
				S9xMainLoop ();
			} else {
				S9xSetSoundMute (TRUE);
				S9xProcessEvents (TRUE);
			}
			S9xProcessEvents (FALSE);
			if (!Settings.Paused)
				S9xSetSoundMute (FALSE);
#endif
		}
	}

	return (0);
}

void S9xAutoSaveSRAM ()
{
	Memory.SaveSRAM (S9xGetFilename (".srm"));
}

void S9xExit ()
{
	S9xSetSoundMute (TRUE);
	Memory.SaveSRAM (S9xGetFilename (".srm"));
//	S9xSaveCheatFile (S9xGetFilename (".cht"));
	Memory.Deinit ();
	S9xDeinitAPU ();

	running = FALSE;
}

Uint16 sfc_key[256];
void S9xInitInputDevices ()
{
	memset(joydown, 0, sizeof(joydown));
	memset(sfc_key, 0, 256);
	sfc_key[QUIT] = SDLK_ESCAPE;
	sfc_key[A_1] = SDLK_d;
	sfc_key[B_1] = SDLK_c;
	sfc_key[X_1] = SDLK_s;
	sfc_key[Y_1] = SDLK_x;
	sfc_key[L_1] = SDLK_a;
	sfc_key[R_1] = SDLK_z;
	sfc_key[START_1] = SDLK_RETURN;
	sfc_key[SELECT_1] = SDLK_SPACE;
	sfc_key[LEFT_1] = SDLK_LEFT;
	sfc_key[RIGHT_1] = SDLK_RIGHT;
	sfc_key[UP_1] = SDLK_UP;
	sfc_key[DOWN_1] = SDLK_DOWN;

	sfc_key[LEFT_2] = SDLK_g;
	sfc_key[RIGHT_2] = SDLK_j;
	sfc_key[UP_2] = SDLK_u;
	sfc_key[DOWN_2] = SDLK_n;
	sfc_key[LU_2] = SDLK_y;
	sfc_key[LD_2] = SDLK_b;
	sfc_key[RU_2] = SDLK_i;
	sfc_key[RD_2] = SDLK_m;

	sfc_key[QUIT] = SDLK_ESCAPE;
	sfc_key[ACCEL] = SDLK_TAB;

	int i = 0;
	char *envp, *j;
	envp = j = getenv ("S9XKEYS");
	if (envp) {
		do {
			if (j = strchr(envp, ','))
				*j = 0;
			if (i == 0) sfc_key[QUIT] = atoi(envp);
			else if (i == 1)  sfc_key[A_1] = atoi(envp);
			else if (i == 2)  sfc_key[B_1] = atoi(envp);
			else if (i == 3)  sfc_key[X_1] = atoi(envp);
			else if (i == 4)  sfc_key[Y_1] = atoi(envp);
			else if (i == 5)  sfc_key[L_1] = atoi(envp);
			else if (i == 6)  sfc_key[R_1] = atoi(envp);
			else if (i == 7)  sfc_key[START_1] = atoi(envp);
			else if (i == 8)  sfc_key[SELECT_1] = atoi(envp);
			else if (i == 9)  sfc_key[LEFT_2] = atoi(envp);
			else if (i == 10) sfc_key[RIGHT_2] = atoi(envp);
			else if (i == 11) sfc_key[UP_2] = atoi(envp);
			else if (i == 12) sfc_key[DOWN_2] = atoi(envp);
			else if (i == 13) sfc_key[LU_2] = atoi(envp);
			else if (i == 14) sfc_key[LD_2] = atoi(envp);
			else if (i == 15) sfc_key[RU_2] = atoi(envp);
			else if (i == 16) sfc_key[RD_2] = atoi(envp);
			envp = j + 1;
			++i;
		} while(j);
	}}

const char *GetHomeDirectory ()
{
	return (getenv ("HOME"));
}

START_EXTERN_C
char* osd_GetPackDir()
{
	static char filename[_MAX_PATH];
	memset(filename, 0, _MAX_PATH);
	
	if(strlen(S9xGetSnapshotDirectory())!=0)
		strcpy (filename, S9xGetSnapshotDirectory());
	else
	{
		char dir [_MAX_DIR + 1];
		char drive [_MAX_DRIVE + 1];
		char name [_MAX_FNAME + 1];
		char ext [_MAX_EXT + 1];
		_splitpath(Memory.ROMFilename, drive, dir, name, ext);
		_makepath(filename, drive, dir, NULL, NULL);
	}

	if(!strncmp((char*)&Memory.ROM [0xffc0], "SUPER POWER LEAG 4   ", 21))
	{
		if (getenv("SPL4PACK"))
			return getenv("SPL4PACK");
		else 
		strcat(filename, "/SPL4-SP7");
	}
	else if(!strncmp((char*)&Memory.ROM [0xffc0], "MOMOTETSU HAPPY      ",21))
	{
		if (getenv("MDHPACK"))
			return getenv("MDHPACK");
		else 
			strcat(filename, "/SMHT-SP7");
	}
	else if(!strncmp((char*)&Memory.ROM [0xffc0], "HU TENGAI MAKYO ZERO ", 21))
	{
		if (getenv("FEOEZPACK"))
			return getenv("FEOEZPACK");
		else 
		strcat(filename, "/FEOEZSP7");
	}
	else if(!strncmp((char*)&Memory.ROM [0xffc0], "JUMP TENGAIMAKYO ZERO",21))
	{
		if (getenv("SJNSPACK"))
			return getenv("SJNSPACK");
		else 
			strcat(filename, "/SJUMPSP7");
	} else strcat(filename, "/MISC-SP7");
	return filename;
}
END_EXTERN_C

const char *S9xGetSnapshotDirectory ()
{
	return ".";
/*
	static char filename [PATH_MAX];
#if 0 // def __GP2X__    
	const char *snapshot;

	if (!(snapshot = getenv ("SNES9X_SNAPSHOT_DIR")) &&
		!(snapshot = getenv ("SNES96_SNAPSHOT_DIR")))
	{
		const char *home = GetHomeDirectory ();
		strcpy (filename, home);
		strcat (filename, SLASH_STR);
		strcat (filename, ".snes96_snapshots");
		mkdir (filename, 0777);
		chown (filename, getuid (), getgid ());
	}
	else
		return (snapshot);
#else

	strcpy(filename, save_path);
#ifdef WIN32
	CreateDirectory(filename, NULL);
#else
	mkdir (filename, 0777);
	chown (filename, getuid (), getgid ());
#endif

#endif

	return (filename);
*/
}

const char *S9xGetFilename (const char *ex)
{
	static char filename [PATH_MAX + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char fname [_MAX_FNAME + 1];
	char ext [_MAX_EXT + 1];

	_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
	strcpy (filename, S9xGetSnapshotDirectory ());
	strcat (filename, SLASH_STR);
	strcat (filename, fname);
	strcat (filename, ex);

	return (filename);
}

const char *S9xGetFilenameInc (const char *e)
{
	static char filename [_MAX_PATH + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char fname [_MAX_FNAME + 1];
	char ext [_MAX_EXT + 1];
	char *ptr;
	struct stat buf;

	if (strlen (S9xGetSnapshotDirectory()))
	{
		_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
		strcpy (filename, S9xGetSnapshotDirectory());
		strcat (filename, "/");
		strcat (filename, fname);
		ptr = filename + strlen (filename);
		strcat (filename, "00/");
		strcat (filename, e);
	}
	else
	{
		_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
		strcat (fname, "00/");
		_makepath (filename, drive, dir, fname, e);
		ptr = strstr (filename, "00/");
	}

	do
	{
		if (++*(ptr + 2) > '9')
		{
			*(ptr + 2) = '0';
			if (++*(ptr + 1) > '9')
			{
				*(ptr + 1) = '0';
				if (++*ptr > '9')
					break;
			}
		}
	} while( stat(filename, &buf) == 0 );

	return (filename);
}

const char *S9xGetROMDirectory ()
{
	const char *roms;
	
	if (!(roms = getenv ("SNES9X_ROM_DIR")) &&
		!(roms = getenv ("SNES96_ROM_DIR")))
		return ("." SLASH_STR "roms");
	else
		return (roms);
}

const char *S9xBasename (const char *f)
{
	const char *p;
	if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
		return (p + 1);

	return (f);
}

#if 0 // #ifndef _ZAURUS
const char *S9xChooseFilename (bool8 read_only)
{
	char def [PATH_MAX + 1];
	char title [PATH_MAX + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char ext [_MAX_EXT + 1];

	_splitpath (Memory.ROMFilename, drive, dir, def, ext);
	strcat (def, ".s96");
	sprintf (title, "%s snapshot filename",
		read_only ? "Select load" : "Choose save");
	const char *filename;

	S9xSetSoundMute (TRUE);
	filename = S9xSelectFilename (def, S9xGetSnapshotDirectory (), "s96", title);
	S9xSetSoundMute (FALSE);
	return (filename);
}
#endif

bool8 S9xOpenSnapshotFile (const char *fname, bool8 read_only, STREAM *file)
{
		if (read_only)
		{
			if (*file = OPEN_STREAM(fname,"rb")) 
				return(TRUE);
		}
		else
		{
			if (*file = OPEN_STREAM(fname,"w+b")) 
				return(TRUE);
		}

		return (FALSE);	
/*
	char filename [PATH_MAX + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char ext [_MAX_EXT + 1];

	_splitpath (fname, drive, dir, filename, ext);

	if (*drive || *dir == '/' ||
		(*dir == '.' && (*(dir + 1) == '/'
		)))
	{
		strcpy (filename, fname);
		if (!*ext)
			strcat (filename, ".s96");
	}
	else
	{
		strcpy (filename, S9xGetSnapshotDirectory ());
		strcat (filename, SLASH_STR);
		strcat (filename, fname);
		if (!*ext)
			strcat (filename, ".s96");
	}
	
#ifdef ZLIB
	if (read_only)
	{
		if ((*file = OPEN_STREAM (filename, "rb")))
			return (TRUE);
	}
	else
	{
		if ((*file = OPEN_STREAM (filename, "wb")))
		{
#ifdef __GP2X__
			chown (filename, getuid (), getgid ());
#endif
			return (TRUE);
		}
	}
#else
	char command [PATH_MAX];
	
	if (read_only)
	{
		sprintf (command, "gzip -d <\"%s\"", filename);
		if (*file = popen (command, "r"))
			return (TRUE);
	}
	else
	{
		sprintf (command, "gzip --best >\"%s\"", filename);
		if (*file = popen (command, "wb"))
			return (TRUE);
	}
#endif
	return (FALSE);
*/
}

void S9xCloseSnapshotFile (STREAM file)
{
#ifdef ZLIB
	CLOSE_STREAM (file);
#else
	pclose (file);
#endif
}

bool8_32 S9xInitUpdate ()
{
	return (TRUE);
}

bool8_32 S9xDeinitUpdate (int Width, int Height, bool8)
{
	register uint32 lp = (xs > 256) ? 16 : 0;

	if (Width > 256 || vga)
		lp *= 2;

//	printf("Render = %d, Skipped = %d.\n", IPPU.RenderThisFrame, IPPU.SkippedFrames);
//	SDL_LockSurface(screen);
	if (Settings.SupportHiRes) {
		if (Width > 256) {
			for (register int i = 0; i < Height; i++) {
				register uint16 *dp16 = (uint16 *)(screen->pixels) + ((i + cl) * xs) + lp;
				register uint32 *sp32 = (uint32 *)(GFX.Screen) + (i << 8) + cs;
				for (register uint32 j = 0; j < 256; j++) {
					*dp16++ = *sp32++;
				}
			}
		} else {
			for (register int i = 0; i < Height; i++) {
				register uint32 *dp32 = (uint32 *)(screen->pixels) + ((i + cl) * xs / 2) + lp;
				register uint32 *sp32 = (uint32 *)(GFX.Screen) + (i << 8) + cs;
				for (register uint32 j = 0; j < 128; j++) {
					*dp32++ = *sp32++;
				}
			}
		}
	}
	if (Settings.DisplayFrameRate)
		S9xDisplayFrameRate ((uint8 *)screen->pixels + 64 + 8 * 640, 640);

//		SDL_UnlockSurface(screen);
	if (ffc < 5) {
		SDL_UpdateRect(screen,0,0,0,0);
		++ffc;
	} else
		SDL_UpdateRect(screen,32,8,256,Height);
//	SDL_Flip(screen);
	return(TRUE);
}

#if 0 // #ifndef _ZAURUS
static unsigned long now ()
{
	static unsigned long seconds_base = 0;
	struct timeval tp;
	gettimeofday (&tp, NULL);
	if (!seconds_base)
	seconds_base = tp.tv_sec;

	return ((tp.tv_sec - seconds_base) * 1000 + tp.tv_usec / 1000);
}
#endif

void _makepath (char *path, const char *, const char *dir,
	const char *fname, const char *ext)
{
	if (dir && *dir)
	{
		strcpy (path, dir);
		strcat (path, "/");
	}
	else
	*path = 0;
	strcat (path, fname);
	if (ext && *ext)
	{
		strcat (path, ".");
		strcat (path, ext);
	}
}

void _splitpath (const char *path, char *drive, char *dir, char *fname,
	char *ext)
{
	*drive = 0;

	char *slash = strrchr (path, '/');
	if (!slash)
		slash = strrchr (path, '\\');

	char *dot = strrchr (path, '.');

	if (dot && slash && dot < slash)
		dot = NULL;

	if (!slash)
	{
		strcpy (dir, "");
		strcpy (fname, path);
		if (dot)
		{
			*(fname + (dot - path)) = 0;
			strcpy (ext, dot + 1);
		}
		else
			strcpy (ext, "");
	}
	else
	{
		strcpy (dir, path);
		*(dir + (slash - path)) = 0;
		strcpy (fname, slash + 1);
		if (dot)
		{
			*(fname + (dot - slash) - 1) = 0;
			strcpy (ext, dot + 1);
		}
		else
			strcpy (ext, "");
	}
}

#if 0 // #ifndef _ZAURUS
void S9xToggleSoundChannel (int c)
{
	if (c == 8)
		so.sound_switch = 255;
	else
		so.sound_switch ^= 1 << c;
	S9xSetSoundControl (so.sound_switch);
}
#endif

static void SoundTrigger ()
{
	if (Settings.APUEnabled && !so.mute_sound)
		S9xProcessSound (NULL);
}

void S9xSyncSpeed(void)
{
	S9xProcessEvents (FALSE);
	if (!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE)
	{
		static Uint64 next = 0;
#if 1
		Uint64 now;
		struct timeval tv;

		gettimeofday(&tv, NULL);
		now = tv.tv_sec;
		now = now * 1000000 + tv.tv_usec;
#else
		Uint64 now = SDL_GetTicks() * 1000;
#endif
		if (next > now) {
			if (IPPU.SkippedFrames == 0) {
#ifdef WIN32
				do {
					CHECK_SOUND ();
					now = SDL_GetTicks() * 1000;
				} while (next > now);
#else
				//now = SDL_GetTicks() * 1000;
				Uint32 delay = next - now;
				//printf("SyncSpeed: %d\n", delay);
				if(delay > Settings.FrameTime)
					usleep(Settings.FrameTime);
				else if(delay > 5000)
					usleep(0);
#endif
			}
			IPPU.RenderThisFrame = TRUE;
			IPPU.SkippedFrames = 0;
		} else {
			if (IPPU.SkippedFrames < mfs) {
				IPPU.SkippedFrames++;
				IPPU.RenderThisFrame = FALSE;
			} else {
				IPPU.RenderThisFrame = TRUE;
				IPPU.SkippedFrames = 0;
				next = now;
			}
		}

		next += Settings.FrameTime;
	} else {
		if (++IPPU.FrameSkip >= (Settings.TurboMode ? Settings.TurboSkipFrames
								: Settings.SkipFrames)) {
			IPPU.FrameSkip = 0;
			IPPU.SkippedFrames = 0;
			IPPU.RenderThisFrame = TRUE;
		} else {
			IPPU.SkippedFrames++;
			IPPU.RenderThisFrame = FALSE;
		}
	}
}

void S9xProcessEvents (bool8_32 block)
{
	uint32 num = 0;
	static bool TURBO = FALSE;
	static int ext = 0;

	SDL_Event event;
	while(SDL_PollEvent(&event)) {
		switch(event.type) {
		case SDL_KEYDOWN:
			keyssnes = SDL_GetKeyState(NULL);

			if (event.key.keysym.sym == sfc_key[QUIT])
				S9xExit();

			if (event.key.keysym.sym == SDLK_0)
				Settings.DisplayFrameRate = !Settings.DisplayFrameRate;
			else if (event.key.keysym.sym == SDLK_1)	PPU.BG_Forced ^= 1;
			else if (event.key.keysym.sym == SDLK_2)	PPU.BG_Forced ^= 2;
			else if (event.key.keysym.sym == SDLK_3)	PPU.BG_Forced ^= 4;
			else if (event.key.keysym.sym == SDLK_4)	PPU.BG_Forced ^= 8;
			else if (event.key.keysym.sym == SDLK_5)	PPU.BG_Forced ^= 16;
			else if (event.key.keysym.sym == SDLK_6)	num = 1;
			else if (event.key.keysym.sym == SDLK_7)	num = 2;
			else if (event.key.keysym.sym == SDLK_8)	num = 3;
			else if (event.key.keysym.sym == SDLK_9)	num = 4;
			else if (event.key.keysym.sym == SDLK_r) {
				if (event.key.keysym.mod & KMOD_SHIFT)
					S9xReset();
			}
			if (num) {
				char fname[256], ext[8];
				sprintf(ext, ".00%d", num - 1);
				strcpy(fname, S9xGetFilename (ext));
				if (event.key.keysym.mod & KMOD_SHIFT)
					S9xFreezeGame (fname);
				else
					S9xLoadSnapshot (fname);
			}
			break;
		case SDL_KEYUP:
			keyssnes = SDL_GetKeyState(NULL);
			break;
#ifdef __GP2X__
		case SDL_JOYBUTTONDOWN:
			joydown[event.jbutton.button] = TRUE;
			if(event.jbutton.button == 0x10) {
				main_volume += 10;
				if(main_volume > 100) main_volume = 100;
				SetVolume();
			} else if(event.jbutton.button == 0x11) {
				main_volume -= 10;
				if(main_volume < 0) main_volume = 0;
				SetVolume();
			}
			break;
		case SDL_JOYBUTTONUP:
			if(joydown[0x12]) {
				joydown[0x12] = FALSE;
				if (joydown[0x0A]) { // Stick + L
					joydown[0x0A] = FALSE;
					switch(event.jbutton.button) {
						case 0x08: // Start
							printf("Load Snapshot: %s\n", snapshot_filename);
							S9xLoadSnapshot (snapshot_filename);
							break;
						case 0x09: // Select
							S9xReset();
							break;
						default:
							joydown[0x12] = joydown[0x0A] = TRUE;
							break;
					}
				} else if (joydown[0x0B]) { // Stick + R
					S9xExit();
				} else { // Stick only
					printf("Button: 0x%02X\n", event.jbutton.button);
					switch(event.jbutton.button) {
						case 0x08: // Start
							printf("Freeze Game: %s\n", snapshot_filename);
							S9xFreezeGame (snapshot_filename);
							break;
						case 0x09: // Select
							Settings.DisplayFrameRate = !Settings.DisplayFrameRate;
							break;
						default:
							joydown[0x12] = TRUE;
							break;
					}
				}
			}
			joydown[event.jbutton.button] = FALSE;
			break;
#endif
		}
		if (keyssnes[sfc_key[ACCEL]] == SDL_PRESSED) {
			if (!TURBO) {
				TURBO = TRUE;
				OldSkipFrame = Settings.SkipFrames;
				Settings.SkipFrames = 10;
			}
		} else {
			if (TURBO) {
				TURBO = FALSE;
				Settings.SkipFrames = OldSkipFrame;
			}
		}
	}
}

static long log2 (long num)
{
	long n = 0;

	while (num >>= 1)
		n++;

	return (n);
}

#if 0 // #ifndef _ZAURUS
static long power (int num, int pow)
{
	long val = num;
	int i;
	
	if (pow == 0)
		return (1);

	for (i = 1; i < pow; i++)
		val *= num;

	return (val);
}
#endif

static int Rates[8] =
{
	0, 8192, 11025, 16500, 22050, 29300, 36600, 44000
};

static int BufferSizes [8] =
{
	0, 256, 256, 256, 512, 512, 1024, 1024
};

void sound_callback(void *data, Uint8 *buff, int len)
{
	SDL_LockAudio();
	if ((main_volume == 0) || Settings.ForcedPause || Settings.Paused || Settings.StopEmulation)
	{
		//ZeroMemory (buff, len);
	}
	else
	{
		int sample_count = len;

		//MK: Sixteenbit was a rendering variable!!
		if (so.sixteen_bit)
			sample_count /= 2;
		//if (main_volume == 128) {
			S9xMixSamplesO ((unsigned char *) buff, sample_count, 0);
		/*} else {
			unsigned char sample[4096];
			S9xMixSamplesO ((unsigned char *) sample, sample_count, 0);
			SDL_MixAudio (buff, sample, len, main_volume);
		}*/

		if (Settings.Mute)
			ZeroMemory (buff, len);
	}
	SDL_UnlockAudio();
}

bool8_32 S9xOpenSoundDevice (int mode, bool8_32 stereo, int buffer_size)
{
	SDL_AudioSpec spec;

	so.sixteen_bit = TRUE;
	so.stereo = stereo;
	so.playback_rate = Rates[mode & 0x07];
	S9xSetPlaybackRate (so.playback_rate);
	so.buffer_size = BufferSizes [mode & 7] * 4;
	if (so.sixteen_bit)
		so.buffer_size *= 2;
	if (so.stereo)
		so.buffer_size *= 2;
	if (so.buffer_size > MAX_BUFFER_SIZE / 4)
		so.buffer_size = MAX_BUFFER_SIZE / 4;

	spec.format = AUDIO_S16;
	spec.channels = (so.stereo ? 2 : 1);
	spec.freq = so.playback_rate;
	spec.samples = so.buffer_size / 4; //1024 * int(spec.freq / 11025);

	spec.callback = sound_callback;
	spec.userdata = NULL;

	if (SDL_OpenAudio(&spec, NULL) < 0) {
		printf("Unable to open audio: %s.\n", SDL_GetError());
		return (FALSE);
	}

	SDL_PauseAudio(0);

	printf ("Rate: %d, Buffer size: %d, 16-bit: %s, Stereo: %s, Encoded: %s\n",
		so.playback_rate, so.buffer_size, so.sixteen_bit ? "yes" : "no",
		so.stereo ? "yes" : "no", so.encoded ? "yes" : "no");

	return (TRUE);
}

void S9xUnixProcessSound (void)
{
}

static uint8 Buf[MAX_BUFFER_SIZE];

#define FIXED_POINT 0x10000
#define FIXED_POINT_SHIFT 16
#define FIXED_POINT_REMAINDER 0xffff

static volatile bool8 block_signal = FALSE;
static volatile bool8 block_generate_sound = FALSE;
static volatile bool8 pending_signal = FALSE;

void S9xGenerateSound () {
	return;
}

uint32 S9xReadJoypad (int which1)
{
	uint32 val=0x80000000;

	if (which1 != 0) return val;
#ifdef __GP2X__
	if (!joydown[0x12]) {
		if (joydown[0x00]) val |= SNES_UP_MASK;
		if (joydown[0x02]) val |= SNES_LEFT_MASK;
		if (joydown[0x04]) val |= SNES_DOWN_MASK;
		if (joydown[0x06]) val |= SNES_RIGHT_MASK;

		if ( (val & (SNES_UP_MASK | SNES_LEFT_MASK | SNES_DOWN_MASK | SNES_RIGHT_MASK)) == 0 ) {
			if (joydown[0x01]) val |= SNES_LEFT_MASK | SNES_UP_MASK;
			if (joydown[0x03]) val |= SNES_LEFT_MASK | SNES_DOWN_MASK;
			if (joydown[0x05]) val |= SNES_RIGHT_MASK | SNES_DOWN_MASK;
			if (joydown[0x07]) val |= SNES_RIGHT_MASK | SNES_UP_MASK;
		}

		if (joydown[0x08]) val |= SNES_START_MASK;
		if (joydown[0x09]) val |= SNES_SELECT_MASK;

		if (joydown[0x0A]) val |= SNES_TL_MASK;
		if (joydown[0x0B]) val |= SNES_TR_MASK;
#if 0
		if (joydown[0x0C]) val |= SNES_A_MASK;
		if (joydown[0x0D]) val |= SNES_B_MASK;
		if (joydown[0x0E]) val |= SNES_X_MASK;
		if (joydown[0x0F]) val |= SNES_Y_MASK;
#elif 1 // original SNES Pad mapping.
		if (joydown[0x0D]) val |= SNES_A_MASK;
		if (joydown[0x0E]) val |= SNES_B_MASK;
		if (joydown[0x0F]) val |= SNES_X_MASK;
		if (joydown[0x0C]) val |= SNES_Y_MASK;
#endif
	}
#endif
	if (keyssnes[sfc_key[L_1]] == SDL_PRESSED)		val |= SNES_TL_MASK;
	if (keyssnes[sfc_key[R_1]] == SDL_PRESSED)		val |= SNES_TR_MASK;
	if (keyssnes[sfc_key[X_1]] == SDL_PRESSED)		val |= SNES_X_MASK;
	if (keyssnes[sfc_key[Y_1]] == SDL_PRESSED)		val |= SNES_Y_MASK;
	if (keyssnes[sfc_key[B_1]] == SDL_PRESSED)		val |= SNES_B_MASK;
	if (keyssnes[sfc_key[A_1]] == SDL_PRESSED)		val |= SNES_A_MASK;
	if (keyssnes[sfc_key[START_1]] == SDL_PRESSED)	val |= SNES_START_MASK;
	if (keyssnes[sfc_key[SELECT_1]] == SDL_PRESSED)	val |= SNES_SELECT_MASK;
	if (keyssnes[sfc_key[UP_1]] == SDL_PRESSED)		val |= SNES_UP_MASK;
	if (keyssnes[sfc_key[DOWN_1]] == SDL_PRESSED)	val |= SNES_DOWN_MASK;
	if (keyssnes[sfc_key[LEFT_1]] == SDL_PRESSED)	val |= SNES_LEFT_MASK;
	if (keyssnes[sfc_key[RIGHT_1]] == SDL_PRESSED)	val |= SNES_RIGHT_MASK;
	if (keyssnes[sfc_key[UP_2]] == SDL_PRESSED)		val |= SNES_UP_MASK;
	if (keyssnes[sfc_key[DOWN_2]] == SDL_PRESSED)	val |= SNES_DOWN_MASK;
	if (keyssnes[sfc_key[LEFT_2]] == SDL_PRESSED)	val |= SNES_LEFT_MASK;
	if (keyssnes[sfc_key[RIGHT_2]] == SDL_PRESSED)	val |= SNES_RIGHT_MASK;
	if (keyssnes[sfc_key[LU_2]] == SDL_PRESSED)	val |= SNES_LEFT_MASK | SNES_UP_MASK;
	if (keyssnes[sfc_key[LD_2]] == SDL_PRESSED)	val |= SNES_LEFT_MASK | SNES_DOWN_MASK;
	if (keyssnes[sfc_key[RU_2]] == SDL_PRESSED)	val |= SNES_RIGHT_MASK | SNES_UP_MASK;
	if (keyssnes[sfc_key[RD_2]] == SDL_PRESSED)	val |= SNES_RIGHT_MASK | SNES_DOWN_MASK;

	return(val);
}

#if 0
void S9xParseConfigFile ()
{
	int i, t = 0;
	char *b, buf[10];
	struct ffblk f;

	set_config_file("SNES9X.CFG");

	if (findfirst("SNES9X.CFG", &f, 0) != 0)
	{
		set_config_int("Graphics", "VideoMode", -1);
		set_config_int("Graphics", "AutoFrameskip", 1);
		set_config_int("Graphics", "Frameskip", 0);
		set_config_int("Graphics", "Shutdown", 1);
		set_config_int("Graphics", "FrameTimePAL", 20000);
		set_config_int("Graphics", "FrameTimeNTSC", 16667);
		set_config_int("Graphics", "Transparency", 0);
		set_config_int("Graphics", "HiColor", 0);
		set_config_int("Graphics", "Hi-ResSupport", 0);
		set_config_int("Graphics", "CPUCycles", 100);
		set_config_int("Graphics", "Scale", 0);
		set_config_int("Graphics", "VSync", 0);
		set_config_int("Sound", "APUEnabled", 1);
		set_config_int("Sound", "SoundPlaybackRate", 4);
		set_config_int("Sound", "Stereo", 1);
		set_config_int("Sound", "SoundBufferSize", 256);
		set_config_int("Sound", "SPCToCPURatio", 2);
		set_config_int("Sound", "Echo", 1);
		set_config_int("Sound", "SampleCaching", 1);
		set_config_int("Sound", "MasterVolume", 1);
		set_config_int("Peripherals", "Mouse", 1);
		set_config_int("Peripherals", "SuperScope", 1);
		set_config_int("Peripherals", "MultiPlayer5", 1);
		set_config_int("Peripherals", "Controller", 0);
		set_config_int("Controllers", "Type", JOY_TYPE_AUTODETECT);
		set_config_string("Controllers", "Button1", "A");
		set_config_string("Controllers", "Button2", "B");
		set_config_string("Controllers", "Button3", "X");
		set_config_string("Controllers", "Button4", "Y");
		set_config_string("Controllers", "Button5", "TL");
		set_config_string("Controllers", "Button6", "TR");
		set_config_string("Controllers", "Button7", "START");
		set_config_string("Controllers", "Button8", "SELECT");
		set_config_string("Controllers", "Button9", "NONE");
		set_config_string("Controllers", "Button10", "NONE");
	}

	mode = get_config_int("Graphics", "VideoMode", -1);
	Settings.SkipFrames = get_config_int("Graphics", "AutoFrameskip", 1);
	if (!Settings.SkipFrames)
		Settings.SkipFrames = get_config_int("Graphics", "Frameskip", AUTO_FRAMERATE);
	else
		Settings.SkipFrames = AUTO_FRAMERATE;
	Settings.ShutdownMaster = get_config_int("Graphics", "Shutdown", TRUE);
	Settings.FrameTimePAL = get_config_int("Graphics", "FrameTimePAL", 20000);
	Settings.FrameTimeNTSC = get_config_int("Graphics", "FrameTimeNTSC", 16667);
	Settings.FrameTime = Settings.FrameTimeNTSC;
	Settings.Transparency = get_config_int("Graphics", "Transparency", FALSE);
	Settings.SixteenBit = get_config_int("Graphics", "HiColor", FALSE);
	Settings.SupportHiRes = get_config_int("Graphics", "Hi-ResSupport", FALSE);
	i = get_config_int("Graphics", "CPUCycles", 100);
	Settings.H_Max = (i * SNES_CYCLES_PER_SCANLINE) / i;
	stretch = get_config_int("Graphics", "Scale", 0);
	_vsync = get_config_int("Graphics", "VSync", 0);

	Settings.APUEnabled = get_config_int("Sound", "APUEnabled", TRUE);
	Settings.SoundPlaybackRate = get_config_int("Sound", "SoundPlaybackRate", 4);
	Settings.Stereo = get_config_int("Sound", "Stereo", TRUE);
	Settings.SoundBufferSize = get_config_int("Sound", "SoundBufferSize", 256);
	Settings.SPCTo65c816Ratio = get_config_int("Sound", "SPCToCPURatio", 2);
	Settings.DisableSoundEcho = get_config_int("Sound", "Echo", TRUE) ? FALSE : TRUE;
	Settings.DisableSampleCaching = get_config_int("Sound", "SampleCaching", TRUE) ? FALSE : TRUE;
	Settings.DisableMasterVolume = get_config_int("Sound", "MasterVolume", TRUE) ? FALSE : TRUE;

	Settings.Mouse = get_config_int("Peripherals", "Mouse", TRUE);
	Settings.SuperScope = get_config_int("Peripherals", "SuperScope", TRUE);
	Settings.MultiPlayer5 = get_config_int("Peripherals", "MultiPlayer5", TRUE);
	Settings.ControllerOption = (uint32)get_config_int("Peripherals", "Controller", SNES_MULTIPLAYER5);

	joy_type = get_config_int("Controllers", "Type", JOY_TYPE_AUTODETECT);
	for (i = 0; i < 10; i++)
	{
		sprintf(buf, "Button%d", i+1);
		b = get_config_string("Controllers", buf, "NONE");
		if (!strcasecmp(b, "A"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_A_MASK;}
		else if (!strcasecmp(b, "B"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_B_MASK;}
		else if (!strcasecmp(b, "X"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_X_MASK;}
		else if (!strcasecmp(b, "Y"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_Y_MASK;}
		else if (!strcasecmp(b, "TL"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_TL_MASK;}
		else if (!strcasecmp(b, "TR"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_TR_MASK;}
		else if (!strcasecmp(b, "START"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_START_MASK;}
		else if (!strcasecmp(b, "SELECT"))
		{JOY_BUTTON_INDEX[t] = i; SNES_BUTTON_MASKS[t++] = SNES_SELECT_MASK;}
	}
}
#endif

static int S9xCompareSDD1IndexEntries (const void *p1, const void *p2)
{
	return (*(uint32 *) p1 - *(uint32 *) p2);
}

void S9xLoadSDD1Data ()
{
	char filename [_MAX_PATH + 1];
	char index [_MAX_PATH + 1];
	char data [_MAX_PATH + 1];
	char patch [_MAX_PATH + 1];

	Memory.FreeSDD1Data ();

	strcpy (filename, S9xGetSnapshotDirectory ());

	if (strncmp (Memory.ROMName, "Star Ocean", 10) == 0)
		strcat (filename, "/socnsdd1");
	else
		strcat (filename, "/sfa2sdd1");

	DIR *dir = opendir (filename);

	index [0] = 0;
	data [0] = 0;
	patch [0] = 0;

	if (dir)
	{
		struct dirent *d;
		
		while ((d = readdir (dir)))
		{
			if (strcasecmp (d->d_name, "SDD1GFX.IDX") == 0)
			{
				strcpy (index, filename);
				strcat (index, "/");
				strcat (index, d->d_name);
			}
			else
				if (strcasecmp (d->d_name, "SDD1GFX.DAT") == 0)
				{
					strcpy (data, filename);
					strcat (data, "/");
					strcat (data, d->d_name);
				}
			if (strcasecmp (d->d_name, "SDD1GFX.PAT") == 0)
			{
				strcpy (patch, filename);
				strcat (patch, "/");
				strcat (patch, d->d_name);
			}
		}
		closedir (dir);

		if (strlen (index) && strlen (data))
		{
			FILE *fs = fopen (index, "rb");
			int len = 0;

			if (fs)
			{
				// Index is stored as a sequence of entries, each entry being
				// 12 bytes consisting of:
				// 4 byte key: (24bit address & 0xfffff * 16) | translated block
				// 4 byte ROM offset
				// 4 byte length
				fseek (fs, 0, SEEK_END);
				len = ftell (fs);
				rewind (fs);
				Memory.SDD1Index = (uint8 *) malloc (len);
				fread (Memory.SDD1Index, 1, len, fs);
				fclose (fs);
				Memory.SDD1Entries = len / 12;

				if (!(fs = fopen (data, "rb")))
				{
					free ((char *) Memory.SDD1Index);
					Memory.SDD1Index = NULL;
					Memory.SDD1Entries = 0;
				}
				else
				{
					fseek (fs, 0, SEEK_END);
					len = ftell (fs);
					rewind (fs);
					Memory.SDD1Data = (uint8 *) malloc (len);
					fread (Memory.SDD1Data, 1, len, fs);
					fclose (fs);

					if (strlen (patch) > 0 &&
					(fs = fopen (patch, "rb")))
					{
						fclose (fs);
					}
#ifdef MSB_FIRST
					// Swap the byte order of the 32-bit value triplets on
					// MSBFirst machines.
					uint8 *ptr = Memory.SDD1Index;
					for (int i = 0; i < Memory.SDD1Entries; i++, ptr += 12)
					{
						SWAP_DWORD ((*(uint32 *) (ptr + 0)));
						SWAP_DWORD ((*(uint32 *) (ptr + 4)));
						SWAP_DWORD ((*(uint32 *) (ptr + 8)));
					}
#endif
					qsort (Memory.SDD1Index, Memory.SDD1Entries, 12,
						S9xCompareSDD1IndexEntries);
				}
			}
		}
		else
		{
			printf ("Decompressed data pack not found in '%s'.\n", filename);
		}
	}
}
