/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
 * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
 *
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: http://www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */
#include <string.h>
#include <ctype.h>
#ifndef WIN32
	#include <unistd.h>
	#include <sys/io.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <signal.h>

#include <SDL.h>

#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "ppu.h"
#include "snapshot.h"
#include "gfx.h"
#include "display.h"
#include "apu.h"
#include "keydef.h"

#ifdef MMUHACK
#include "squidgehack.h"
#endif

#define COUNT(a) (sizeof(a) / sizeof(a[0]))

#define bool8_32 bool8

SDL_Surface *screen;
SDL_Surface *gfxscreen;
uint16 *RGBconvert;
extern uint32 xs, ys, cl, cs;

int S9xMinCommandLineArgs ()
{
    return (2);
}

void S9xGraphicsMode ()
{
}

void S9xTextMode ()
{
}

extern uint8 *keyssnes;
extern bool8 sdl_init;
void S9xInitDisplay (int /*argc*/, char ** /*argv*/)
{
#ifdef MMUHACK
	if(!sdl_init) {
	      hackpgtable();
	}
#endif
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | (Settings.NextAPUEnabled ? SDL_INIT_AUDIO : 0)) < 0 ) 
	{
		printf("Could not initialize SDL(%s)\n", SDL_GetError());
		S9xExit();
	}
	sdl_init = 1;
	SDL_ShowCursor(SDL_DISABLE);
	keyssnes = SDL_GetKeyState(NULL);

	// Check for joystick
	if(SDL_NumJoysticks()>0)
	{
		SDL_Joystick *joy = SDL_JoystickOpen(0);
  		
  		if(joy)
  		{
    		printf("Opened Joystick 0\n");
    		printf("Name: %s\n", SDL_JoystickName(0));
    		printf("Number of Axes: %d\n", SDL_JoystickNumAxes(joy));
    		printf("Number of Buttons: %d\n", SDL_JoystickNumButtons(joy));
    		printf("Number of Balls: %d\n", SDL_JoystickNumBalls(joy));
	  	}
  		else
  		{
    		printf("Couldn't open Joystick 0\n");
    	}
	}

	printf("Resolution : %dx%d\n", xs, ys);
#ifdef __GP2X__
	screen = SDL_SetVideoMode(320, 240, 16, SDL_FULLSCREEN);
#else
	screen = SDL_SetVideoMode(320, 240, 16, SDL_HWSURFACE|SDL_DOUBLEBUF/*|SDL_FULLSCREEN*/);
#endif
	if (screen == NULL)
	{
		printf("Couldn't set video mode: %s\n", SDL_GetError());
		S9xExit();
	}
	printf("Screen settings(HiRes = %d, MustLock = %d).\n", Settings.SupportHiRes, SDL_MUSTLOCK(screen));
	SDL_ShowCursor(SDL_DISABLE);
	if (Settings.SupportHiRes) {
		gfxscreen = NULL;
		GFX.Screen = (uint8 *)malloc(512 * 480 * 2);
		GFX.Pitch = 512 * 2;
	} else {
		gfxscreen = NULL;
		GFX.Screen = (uint8 *)screen->pixels + 64 + 8 * 640;
		GFX.Pitch = 320 * 2;
	}
	GFX.SubScreen = (uint8 *)malloc(512 * 480 * 2);
	GFX.ZBuffer = (uint8 *)malloc(512 * 480 * 2);
	GFX.SubZBuffer = (uint8 *)malloc(512 * 480 * 2);

	RGBconvert = (uint16 *)malloc(65536 * 2);
	if (!RGBconvert)
	{
//		OutOfMemory();
		S9xExit();
	}
	for (uint32 i = 0; i < 65536; i++) 
		((uint16 *)(RGBconvert))[i] = ((i >> 11) << 10) | ((((i >> 5) & 63) >> 1) << 5) | (i & 31);
}

void S9xDeinitDisplay ()
{
	if(gfxscreen != NULL)
		SDL_FreeSurface(gfxscreen);
	SDL_FreeSurface(screen);
	free(RGBconvert);
	if (Settings.SupportHiRes)
		free(GFX.Screen);
	free(GFX.SubScreen);
	free(GFX.ZBuffer);
	free(GFX.SubZBuffer);
}

void S9xSetPalette ()
{
}

void S9xSetTitle (const char * /*title*/)
{
}

const char *S9xSelectFilename (const char *def, const char *dir1,
			    const char *ext1, const char *title)
{
    static char path [PATH_MAX];
    char buffer [PATH_MAX];
    
    S9xTextMode ();
    printf ("\n%s (default: %s): ", title, def);
    fflush (stdout);
    if (fgets (buffer, sizeof (buffer) - 1, stdin))
    {
	char *p = buffer;
	while (isspace (*p) || *p == '\n')
	    p++;
	if (!*p)
	{
	    strcpy (buffer, def);
	    p = buffer;
	}

	char *q = strrchr (p, '\n');
	if (q)
	    *q = 0;

	char fname [PATH_MAX];
	char drive [_MAX_DRIVE];
	char dir [_MAX_DIR];
	char ext [_MAX_EXT];

	_splitpath (p, drive, dir, fname, ext);
	_makepath (path, drive, *dir ? dir : dir1, fname, *ext ? ext : ext1);
	S9xGraphicsMode ();
	return (path);
    }
    S9xGraphicsMode ();
    return (NULL);
}

void S9xParseDisplayArg (char **argv, int &ind, int)
{
/*
    if (strcasecmp (argv [ind], "-scale") == 0 ||
	strcasecmp (argv [ind], "-sc") == 0)
	stretch = TRUE;
    else
    if (strcasecmp (argv [ind], "-y") == 0 ||
	strcasecmp (argv [ind], "-interpolation") == 0)
    {
	interpolation = TRUE;
	Settings.SixteenBit = TRUE;
	Settings.SupportHiRes = TRUE;
	Settings.Transparency = TRUE;
    }
    else
*/
	S9xUsage ();
}

void S9xExtraUsage ()
{
}

bool8 S9xReadMousePosition (int /* which1 */, int &/* x */, int & /* y */,
			    uint32 & /* buttons */)
{
    return (FALSE);
}

bool8 S9xReadSuperScopePosition (int & /* x */, int & /* y */, 
				 uint32 & /* buttons */)
{
    return (FALSE);
}

void S9xMessage (int /* type */, int /* number */, const char *message)
{
    printf ("Message : %s\n", message);
}

bool JustifierOffscreen()
{
  return false;
}

void JustifierButtons(uint32& justifiers)
{
}
